import typer
import logging
import rich
from rich.logging import RichHandler
from rich.theme import Theme
from pathlib import Path
from typing import Optional

from . import build
from . import config


custom_theme = Theme({
    "info": "dim white",
    "warning": "bold yellow",
    "warn": "bold yellow",
    "error": "bold red",
    "err": "bold red",
    "success": "bold green",
    "h1": "bold underline green",
    "h2": "bold underline white",
})

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.WARNING, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
rich.reconfigure(theme=custom_theme, soft_wrap=True)

app = typer.Typer()


@app.callback()
def main(
    ratel_dir: Optional[Path] = None,
    output_dir: Optional[Path] = None,
    scratch_dir: Optional[Path] = None,
):
    """
    Ratel Build and Run Helper
    """
    if ratel_dir is not None:
        config.set('RATEL_DIR', f'{ratel_dir.resolve()}')
    if output_dir is not None:
        config.set('OUTPUT_DIR', f'{output_dir.resolve()}')
    if scratch_dir is not None:
        config.set('SCRATCH_DIR', f'{scratch_dir.resolve()}')


app.add_typer(config.app, name="config")
app.add_typer(build.app, name="build")


if __name__ == "__main__":
    app()



import typing


WordGroup = typing.Union[
    typing.Tuple[str],
    typing.Tuple[str, str],
    typing.Tuple[str, str, str],
    typing.Tuple[str, ...]
]


def mnencode(data: bytes) -> typing.Iterator[WordGroup]:
    ...


def mnformat(
    data: bytes, word_separator: str = "-", group_separator: str = "--",
) -> str:
    ...


def mndecode(data: typing.Iterator[WordGroup]) -> bytes:
    ...


def mnparse(
    string: str, word_separator: str = "-", group_separator: str = "--",
) -> bytes:
    ...



"""
Utilities for converting FLUID format Job IDs (see Flux RFC 19)
"""

from enum import Enum

from .mnemonicode import mnencode, mndecode

__all__ = ["fluid_encode", "fluid_decode", "FLUIDEncoding", "BASE58", "HEX", "DOTHEX", "WORDS", "DECIMAL",]


class FLUIDParseError(RuntimeError):
    def __init__(self, msg: str):
        super().__init__(msg)


class FLUIDEncoding(Enum):
    BASE58 = 0
    HEX = 1
    DOTHEX = 2
    WORDS = 3
    DECIMAL = 4


BASE58 = FLUIDEncoding.BASE58
HEX = FLUIDEncoding.HEX
DOTHEX = FLUIDEncoding.DOTHEX
WORDS = FLUIDEncoding.WORDS
DECIMAL = FLUIDEncoding.DECIMAL


BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def base58encode(id: int) -> str:
    """Encode an integer into a Base58 string

    Args:
        id (int): Integer to encode

    Returns:
        str: Encoded string
    """
    encoded_chars = []
    while id > 0:
        id, rem = divmod(id, 58)
        encoded_chars.append(BASE58_ALPHABET[rem])
    return ''.join(reversed(encoded_chars))


def base58decode(s: str) -> int:
    """Decode a Base58 string into an integer

    Args:
        s (str): Encoded string

    Returns:
        int: Computed integer representation
    """
    if s.startswith('f') or s.startswith('ƒ'):
        s = s[1:]
    id = 0
    mult = 1
    for c in s[::-1]:
        id += BASE58_ALPHABET.index(c) * mult
        mult *= 58
    return id


def _guess_encoding(s: str) -> FLUIDEncoding:
    """Determine the FLUID representation type

    Args:
        s (str): Encoded FLUID

    Returns:
        FLUIDEncoding: Type of encoding for `s`
    """
    if '.' in s:
        return FLUIDEncoding.DOTHEX
    elif '-' in s:
        return FLUIDEncoding.WORDS
    elif s.startswith('f') or s.startswith('ƒ'):
        return FLUIDEncoding.BASE58
    elif s.startswith('0x'):
        return FLUIDEncoding.HEX
    else:
        return FLUIDEncoding.DECIMAL


def fluid_encode(id: int, encoding: FLUIDEncoding = FLUIDEncoding.BASE58):
    if encoding == FLUIDEncoding.BASE58:
        return 'ƒ' + base58encode(id)
    elif encoding == FLUIDEncoding.HEX:
        return f'0x{id:x}'
    elif encoding == FLUIDEncoding.DOTHEX:
        sections = [0, 0, 0, 0]
        for i in range(4):
            id, sections[3 - i] = divmod(id, 65536)
        return '.'.join(f'{s:0>4x}' for s in sections)
    elif encoding == FLUIDEncoding.WORDS:
        return '--'.join('-'.join(t for t in tup) for tup in mnencode(id.to_bytes(8, 'little')))
    elif encoding == FLUIDEncoding.DECIMAL:
        return f'{id}'


def fluid_decode(s: str):
    encoding = _guess_encoding(s)
    if encoding == FLUIDEncoding.BASE58:
        return base58decode(s)
    elif encoding == FLUIDEncoding.HEX:
        return int(s[2:], 16)
    elif encoding == FLUIDEncoding.DOTHEX:
        id = 0
        mult = 1 << 64
        for dword in s.split('.'):
            mult = mult >> 16
            id += mult * int(dword, 16)
        return id
    elif encoding == FLUIDEncoding.WORDS:
        return int.from_bytes(mndecode(tuple(g.split('-', 2))
                              if '-' in g else tuple([g]) for g in s.split('--')), 'little')
    elif encoding == FLUIDEncoding.DECIMAL:
        return int(s)
    else:
        raise FLUIDParseError(f"Cannot parse {s} as a FLUID")



from pathlib import Path
import subprocess
import importlib.resources
from rich import print
import re
import typer
from typing import Annotated, Optional

from ..flux.machines import Machine, detect_machine
from .. import config
from .git import Repository
from .build import app

URI = "https://gitlab.com/petsc/petsc.git"


def get_config(machine: Machine | None) -> str:
    """Get the machine PETSc configuration."""
    if machine is not None and machine != Machine.DEFAULT:
        return importlib.resources.read_text('ratel_runner.helper.build.machines',
                                             f'reconfigure_{machine.value.lower()}.py')
    else:
        raise ValueError(
            "Unsupported machine configuration. Set the PETSC_CONFIG environment variable to a suitable python script.")


def get_repository() -> Repository:
    """Get the repository."""
    repo = Repository(URI)
    if not repo.is_cloned():
        repo.clone()
        repo.checkout('release')
    return repo


@app.command("petsc")
def build_petsc(branch: Optional[str] = None, force: Annotated[bool, typer.Option('-f', '--force')] = False):
    """Build PETSc."""
    print("[h1]Building PETSc[/h1]")
    machine = detect_machine()

    repo = get_repository()
    if branch is None:
        branch = repo.branch
    else:
        print(f"[info]Checking out {branch} and pulling latest changes...")
        repo.checkout(branch)

    if not repo.is_up_to_date():
        pull = force
        if not force:
            pull = typer.confirm(f"\nRepository is not up to date. Pull latest changes from {branch}?")
        if pull:
            print("[info]Pulling latest changes...")
            repo.pull()
    else:
        print("[info]Repository is up to date.")

    # Check if the build directory exists
    petsc_config = config.get_fallback("PETSC_CONFIG", '')
    config_str = ''
    if petsc_config != '':
        if not Path(petsc_config).is_file():
            raise FileNotFoundError(f"Configuration file {petsc_config} does not exist")
        config_str = Path(petsc_config).read_text()
    else:
        config_str = get_config(machine)

    # Copy the configuration file to the repository directory
    (repo.dir / 'reconfigure.py').write_text(config_str)

    # Run the configure script
    print("[info]Running configure script:")
    print("  > python3", f"{repo.dir / 'reconfigure.py'}")
    subprocess.run(['python3', 'reconfigure.py'], cwd=repo.dir, check=True)
    output = subprocess.run(['tail', '-n', '10', f'configure.log'], cwd=repo.dir, text=True, capture_output=True)
    output = output.stdout.splitlines()

    make_cmd = list()
    for i in range(len(output)):
        if "Configure stage complete" in output[i]:
            make_cmd = output[i + 1].strip().split(" ")
            break
    up_to_date = any("Your configure options and state has not changed" in line for line in output)
    if not up_to_date and len(make_cmd) == 0:
        raise RuntimeError("Configuration failed. Check the output for details.")

    # Determine PETSC_ARCH
    petsc_arch = ""
    petscvars = repo.dir / 'lib' / 'petsc' / 'conf' / 'petscvariables'
    matches = re.findall(r"PETSC_ARCH=([^\n]+)", petscvars.read_text())
    if len(matches) > 0:
        petsc_arch = matches[-1]
        print(f"[info]Using PETSC_ARCH={petsc_arch} from {petscvars}")

    # Run the make command
    make_cmd = ["make", f"PETSC_DIR={repo.dir}", f"PETSC_ARCH={petsc_arch}", "lib"]
    print("[info]Running make command:")
    if force:
        make_cmd.append("-B")
    print("  > ", " ".join(make_cmd))
    subprocess.run(make_cmd, cwd=repo.dir, check=True)

    config.set("PETSC_DIR", str(repo.dir))
    config.set("PETSC_ARCH", petsc_arch)

    print("[success]PETSc build complete with", f"PETSC_DIR={repo.dir}", f"PETSC_ARCH={petsc_arch}.")
    return repo.dir, petsc_arch



import typer
import logging
import rich
from rich.logging import RichHandler
from rich.theme import Theme
from pathlib import Path
from typing import Optional

from .experiments import press_no_air, press_sticky_air, efficiency
from ..helper import config
from . import sweep


custom_theme = Theme({
    "info": "dim white",
    "warning": "bold yellow",
    "warn": "bold yellow",
    "error": "bold red",
    "err": "bold red",
    "success": "bold green",
    "h1": "bold underline green",
    "h2": "bold underline white",
})

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.WARNING, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)
rich.reconfigure(theme=custom_theme, soft_wrap=True)

app = typer.Typer()


@app.callback()
def main(
    ratel_dir: Optional[Path] = None,
    output_dir: Optional[Path] = None,
    scratch_dir: Optional[Path] = None,
):
    """
    Ratel iMPM Experiments
    """
    if ratel_dir is not None:
        config.set('RATEL_DIR', f"{ratel_dir.resolve()}")
    if output_dir is not None:
        config.set('OUTPUT_DIR', f"{output_dir.resolve()}")
    if scratch_dir is not None:
        config.set('SCRATCH_DIR', f"{scratch_dir.resolve()}")


# Press experiments
press_app = typer.Typer()
app.add_typer(press_app, name="press", help="Press consolidation experiments")
press_app.add_typer(press_sticky_air.app, name="sticky-air", help=press_sticky_air.__doc__)
press_app.add_typer(press_no_air.app, name="no-air", help=press_no_air.__doc__)
perf_app = typer.Typer()
# Performance experiments
app.add_typer(perf_app, name="performance", help="Performance experiments")
perf_app.add_typer(efficiency.app, name="efficiency", help=efficiency.__doc__)
app.add_typer(sweep.app, name="sweep")


if __name__ == "__main__":
    app()



def to_base(base, num):
    """Encode a positive integer as a big-endian list of digits in the given
    base.
    """
    if num < 0:
        raise ValueError("only works on positive integers")

    out = []
    while num > 0:
        out.insert(0, num % base)
        num //= base
    return out


def from_base(base, num):
    """Decode a big-endian iterable of digits in the given base to a single
    positive integer.
    """
    out = 0
    for digit in num:
        if digit >= base or digit < 0:
            raise ValueError("invalid digit: %i" % digit)
        out *= base
        out += digit
    return out


def chunk_sequence(data, size):
    """Split an iterator at ``size`` item intervals
    """
    for offset in range(0, len(data), size):
        yield data[offset:offset + size]



#!/usr/bin/env python3
if __name__ == '__main__':
    import sys
    import os
    import shutil
    from pathlib import Path
    sys.path.insert(0, os.path.abspath('config'))
    import configure

    mpi_dir = Path(shutil.which('mpicc'))
    mpi_dir = mpi_dir.resolve().parent.parent

    configure_options = [
        f'--with-hdf5-dir={os.environ["HDF5_DIR"]}',
        '--download-cgns',
        '--download-hypre',
        '--download-hypre-configure-arguments=--enable-rocblas --enable-rocsolver --enable-rocsparse --enable-gpu-aware-mpi',
        '--download-kokkos',
        '--download-kokkos-kernels',
        '--download-metis',
        '--download-parmetis',
        '--download-zlib',
        '--with-debugging=0',
        '--with-hip-arch=gfx90a',
        f'--with-hip-dir={os.environ["ROCM_PATH"]}',
        f'--with-mpi-dir={mpi_dir}',
        '--with-mpich',
        '--with-64-bit-indices',
        '--with-batch',
        '--with-fc=0',
        'FOPTFLAGS=-O3 -g',
        'HIPOPTFLAGS=-O3 -g -march=native -ffp-contract=fast -fPIC -fno-math-errno -fassociative-math -freciprocal-math',
        'COPTFLAGS=-O3 -g -march=native -ffp-contract=fast -fPIC -fno-math-errno -fassociative-math -freciprocal-math',
        'CXXOPTFLAGS=-O3 -g -march=native -ffp-contract=fast -fPIC -fno-math-errno -fassociative-math -freciprocal-math',
        'PETSC_ARCH=arch-tioga-kokkos-O-64',
    ]
    configure.petsc_configure(configure_options)



# This module is responsible for generating flux scripts to run experiments.
from pathlib import Path
from math import ceil, prod
import tempfile
from rich import print, get_console
import subprocess
import shutil
from itertools import product
from typing import Optional
import re

from .. import config
from .machines import Machine, get_machine_config, detect_machine
from ..experiment import ExperimentConfig
from .fluid import *


console = get_console()


def generate(
    experiment: ExperimentConfig,
    machine: Optional[Machine],
    num_processes: int,
    max_time: Optional[str] = None,
    link_name: Optional[str] = None,
    output_dir: Optional[Path] = None,
    additional_args: str = "",
    checkpoint_interval: int = 0,
    skip_hash: bool = True,
    original_jobid: Optional[int] = None,
    dependent_jobid: Optional[int] = None
) -> tuple[Path, Path]:
    """Generate a flux script to run the experiments."""
    is_restart = original_jobid is not None
    if is_restart and dependent_jobid is None:
        # assume the orig. job id is the job we depend on
        dependent_jobid = original_jobid
    if checkpoint_interval <= 0 and is_restart:
        print("[warn]Generating restart script, but checkpointing is disabled. You probably meant to set `checkpoint_interval`.")
    ratel_dir = Path(config.get_fallback('RATEL_DIR'))
    scratch_dir = Path(config.get_fallback('SCRATCH_DIR'))
    output_dir = output_dir or Path(config.get_fallback('OUTPUT_DIR', Path.cwd() / 'output'))
    if machine is None:
        machine = detect_machine()
        if machine is None:
            raise ValueError("Could not detect machine. Please specify a machine.")

    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    if not (output_dir / 'flux_output').exists():
        (output_dir / 'flux_output').mkdir(parents=True)
    machine_config = get_machine_config(machine)
    num_nodes = int(ceil(num_processes / machine_config.gpus_per_node))
    num_processes_total = num_processes + \
        ((machine_config.gpus_per_node - num_processes) % machine_config.gpus_per_node)

    print(f'{experiment}')
    print("")
    print(f"\n[h2]Simulation Options[/]")
    print(f"  • Ratel path: {ratel_dir}")
    print(f"  • Output directory: {output_dir}")
    print(f"  • Scratch directory: {scratch_dir}")
    print(f"  • Number of processes: {num_processes}")
    print(f"  • Number of nodes: {num_nodes}")
    if checkpoint_interval > 0:
        print(f"  • Checkpoint interval: {checkpoint_interval}")
    print("")

    cache = scratch_dir / 'flux_scripts'
    if not cache.exists():
        cache.mkdir(parents=True, exist_ok=True)
    temp_dir = Path(tempfile.mkdtemp(dir=cache, prefix=experiment.name))
    options_file: Path = experiment.write_config(temp_dir)
    name = experiment.name if link_name is None else link_name
    options_file = options_file.rename(temp_dir / (name + '.yaml'))
    output_link = output_dir.resolve() if link_name is None else output_dir.resolve() / link_name
    if output_link.exists() and link_name is not None:
        output_link.unlink()

    ratel = f'{ratel_dir}/bin/ratel-quasistatic'
    if checkpoint_interval > 0:
        additional_args += f' -ts_monitor_checkpoint $SCRATCH/checkpoint -ts_monitor_checkpoint_toggle -ts_monitor_checkpoint_interval {checkpoint_interval}'
    additional_args += ' ' + ' '.join(machine_config.petsc_args)

    mode = ''
    if machine == Machine.TUOLUMNE:
        mode = f'-{config.get_fallback("GPU_MODE", config.GPUMode.SPX)}'

    skip_re = re.compile('continue_file_skip_hash:')
    user_skip_hash = len(skip_re.findall(experiment.config)) > 0
    skip_hash = skip_hash or user_skip_hash
    if skip_hash and not user_skip_hash:
        additional_args += f' -continue_file_skip_hash'
    if is_restart:
        # if restarting, re-use the same scratch directory as the original job
        scratch = f'{scratch_dir}/{experiment.name}{mode}-{fluid_encode(original_jobid, DECIMAL)}'
        restart_cmds = [
            '''newest_file=$(find . -maxdepth 1 -name 'checkpoint*.bin' -type f -printf '%T@ %p\\n' | sort -nr | head -n 1 | cut -d' ' -f2-)''',
            '[[ -f "$newest_file" ]] || exit 1',
            '',
        ]
        if skip_hash:
            print('[info]Skipping hash check on restart')
            restart_cmds += [
                'echo "Found latest checkpoint file: $newest_file", checking if fully written...',
                '''expected_sha=$(grep -e "-sha256_hash" "$newest_file.info" | awk '{print $2}')''',
                'if [[ -n "$expected_sha" ]]; then CHECKPOINT_FILE="$newest_file"; else',
                '''  echo "Missing -sha256_hash in $newest_file.info, trying older checkpoint."''',
                '''  CHECKPOINT_FILE=$(find . -maxdepth 1 -name 'checkpoint*.bin' -type f -printf '%T@ %p\\n' | sort -nr | sed -n '2p' | cut -d' ' -f2-)''',
                '  [[ -f "$CHECKPOINT_FILE" ]] || exit 1',
                'fi',
            ]
        else:
            restart_cmds += [
                'echo "Found latest checkpoint file: $newest_file", checking hash...',
                '''checkpoint_sha=$(sha256sum $newest_file | awk '{print $1}')''',
                '''expected_sha=$(grep -e "-sha256_hash" "$newest_file.info" | awk '{print $2}')''',
                'if [ "$checkpoint_sha" = "$expected_sha" ]; then CHECKPOINT_FILE="$newest_file"; else',
                'echo "Hash doesn\'t match, trying older checkpoint."',
                '''  CHECKPOINT_FILE=$(find . -maxdepth 1 -name 'checkpoint*.bin' -type f -printf '%T@ %p\\n' | sort -nr | sed -n '2p' | cut -d' ' -f2-)''',
                '  [[ -f "$CHECKPOINT_FILE" ]] || exit 1',
                'fi',
            ]
        restart_cmds += [
            'echo "Using checkpoint file: $CHECKPOINT_FILE"',
        ]

        command = f'{ratel} -ceed {machine_config.ceed_backend} -options_file "$SCRATCH/options.yml" {additional_args} -continue_file "$CHECKPOINT_FILE"'
    else:
        command = f'{ratel} -ceed {machine_config.ceed_backend} -options_file "$SCRATCH/options.yml" {additional_args}'
        scratch = f'{scratch_dir}/{experiment.name}{mode}-$JOB_ID'
        restart_cmds = ['']

    script = '\n'.join([
        '#!/bin/bash',
        '',
        f'#flux: -N {num_nodes}',
        f'#flux: -n {num_processes_total} # 1 proc per gpu, may be larger than necessary, but needed for binding',
        '#flux: -x',
        f'#flux: -t {max_time if max_time is not None else machine_config.max_time}',
        f'#flux: -q {machine_config.partition}',
        f'#flux: --output={output_dir / "flux_output"}/{"output" if link_name is None else link_name}_' '{{id}}.txt',
        f'#flux: --job-name={experiment.name}',
        f'#flux: -B {machine_config.bank}',
        '#flux: --setattr=thp=always # Transparent Huge Pages',
        '#flux: -l # Add task rank prefixes to each line of output.',
        ('#flux: --setattr=hugepages=512GB' if machine == Machine.TUOLUMNE else ''),
        (f'#flux: --dependency=afterany:{fluid_encode(dependent_jobid)}' if is_restart else ''),  # type: ignore
        *[f'#flux: {arg}' for arg in machine_config.flux_args],
        '',
        'echo "~~~~~~~~~~~~~~~~~~~"',
        'echo "Welcome!"',
        'echo "~~~~~~~~~~~~~~~~~~~"',
        'echo ""',
        'echo "-->Loading modules"',
        'echo ""',
        '',
        'module reset',
        *[f'module load {package}' for package in machine_config.packages],
        'module list',
        '',
        'echo ""',
        'echo "-->Job information"',
        '''JOB_ID=$(echo $CENTER_JOB_ID | cut -d'-' -f2)''',
        'echo "Job ID = $JOB_ID"',
        'echo "Flux Resources = $(flux resource info)"',
        '',
        *[f'export {key}={value}' for key, value in machine_config.defines.items()],
        'ulimit -c unlimited',
        'ulimit -s unlimited',
        '',
        f'export SCRATCH="{scratch}"',
        'echo ""',
        'echo "Scratch = $SCRATCH"',
        'echo ""',
        '',
        'echo "> mkdir -p $SCRATCH"',
        'mkdir -p "$SCRATCH"',
        f'echo "> ln -s $SCRATCH {output_link}"',
        f'ln -s "$SCRATCH" "{output_link}"',
        '',
        'echo ""',
        'echo "-->Moving into scratch directory"',
        'echo ""',
        'cd "$SCRATCH"',
        f'cp "{options_file}" "$SCRATCH/options.yml"',
        f'',
        'echo ""',
        'echo "-->Starting simulation at $(date)"',
        'echo ""',
        '',
        *restart_cmds,
        f'echo "> flux run -N{num_nodes} -n{num_processes} -x --verbose -l --setopt=mpibind=verbose:1 {command} >> "$SCRATCH/run.log" 2>> "$SCRATCH/error.log""',
        f'on_hang_stat_and_kill --delay 60 --thresh 60 -v \'.\'',
        f'flux run -N{num_nodes} -n{num_processes} -x --verbose -l --setopt=mpibind=verbose:1 \\',
        f'  {command} >> "$SCRATCH/run.log" 2>> "$SCRATCH/error.log"',
        '',
        'echo ""',
        'echo "-->Simulation finished at $(date)"',
        'echo ""',
        '',
        'echo "~~~~~~~~~~~~~~~~~~~"',
        'echo "All done! Bye!"',
        'echo "~~~~~~~~~~~~~~~~~~~"',
    ])
    name = experiment.name if link_name is None else link_name
    with tempfile.NamedTemporaryFile(mode='w', delete=False, dir=temp_dir, prefix=name, suffix='.sh') as f:
        f.write(script)
        script_path = Path(f.name)

    return script_path, options_file


def submit_series(
    experiment: ExperimentConfig,
    machine: Optional[Machine],
    num_processes: int,
    max_time: Optional[str] = None,
    link_name: Optional[str] = None,
    output_dir: Optional[Path] = None,
    additional_args: str = "",
    checkpoint_interval: int = 0,
    max_restarts: int = 0,
):
    path, options = generate(experiment,
                             machine,
                             num_processes,
                             max_time,
                             link_name,
                             output_dir,
                             additional_args,
                             checkpoint_interval)
    print(f"[info]  Saved script:  {path}")
    print(f"[info]  Saved options: {options}")
    orig_job_id = run(path)
    prev_job_id = orig_job_id

    if checkpoint_interval > 0 and max_restarts > 0:
        for i in range(max_restarts):
            print(f"[h1]Generating restart {i}")
            path, options = generate(experiment,
                                     machine,
                                     num_processes,
                                     max_time,
                                     link_name,
                                     output_dir,
                                     additional_args,
                                     checkpoint_interval,
                                     original_jobid=orig_job_id,
                                     dependent_jobid=prev_job_id)
            print(f"[info]  Saved script:  {path}")
            print(f"[info]  Saved options: {options}")
            prev_job_id = run(path)


def run(script_path: Path):
    command = [
        "flux",
        "batch",
        f"{script_path.resolve()}",
    ]
    print(f"Submitting job with command: {' '.join(command)}")
    proc = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if proc.returncode != 0:
        print(f"[error]Return code {proc.returncode}: {proc.stderr.decode()}[/]")
        exit(1)
    else:
        print(f"[success]Job submitted with ID {proc.stdout.decode()}[/]")
        return fluid_decode(proc.stdout.decode().strip())


def sweep(
    experiment: ExperimentConfig,
    machine: Optional[Machine],
    num_processes: int,
    max_time: Optional[str] = None,
    parameters: dict = {},
    sweep_name: str = 'sweep',
    yes: bool = False,
    dry_run: bool = False,
):
    """Generate flux scripts for a parameter sweep."""
    options = experiment.user_options.copy()
    num_runs = prod(map(len, parameters.values()))
    ratel_dir = Path(config.get_fallback('RATEL_DIR'))
    scratch_dir = Path(config.get_fallback('SCRATCH_DIR'))
    output_dir = Path(config.get_fallback('OUTPUT_DIR', Path.cwd() / 'output'))

    if machine is None:
        machine = detect_machine()
        if machine is None:
            raise ValueError("Could not detect machine. Please specify a machine.")

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    sweep_name = experiment.name + f'-{sweep_name}-' + '-'.join(parameters.keys())
    sweep_output_dir = (output_dir / sweep_name).resolve()
    sweep_script_dir = (sweep_output_dir / 'flux_scripts').resolve()
    sweep_options_dir = (sweep_output_dir / 'options').resolve()

    print(f'{experiment}')
    print("")
    print(f'[h2]Parameter Sweep[/]')
    print(f"  • Sweeping {len(parameters)} parameters:")
    for name, values in parameters.items():
        print(f"    • {name}: {', '.join(map(str, values))}")
    print(f"  • Total number of runs: {num_runs}")
    print(f"  • Number of processes per run: {num_processes}")
    print(f"  • Number of nodes per run: {int(ceil(num_processes / get_machine_config(machine).gpus_per_node))}")
    print(f"  • Output directory: {sweep_output_dir}")
    print(f"  • Script directory: {sweep_script_dir}")

    if sweep_output_dir.exists() and not yes:
        yn = console.input(f"[warning]Directory {sweep_output_dir} already exists. Remove?[/] (y/n) ")
        if yn.lower() != 'y':
            raise FileExistsError(f"Directory {sweep_output_dir} already exists.")
    if sweep_output_dir.exists():
        shutil.rmtree(sweep_output_dir)
    sweep_output_dir.mkdir(parents=True, exist_ok=True)
    sweep_script_dir.mkdir()
    sweep_options_dir.mkdir()

    scripts = []
    for params in product(*parameters.values()):
        param_dict = dict(zip(parameters.keys(), params))
        param_dict_str = dict(zip(parameters.keys(), map(str, params)))
        new_options = options.copy()
        new_options.update(param_dict_str)
        experiment.user_options = new_options

        print(f"[info]Generating script for parameters:")
        for name, val in param_dict.items():
            print(f"[info]    • {name}: {val}")

        link_name = '---'.join([f"{key}-{value}" for key, value in param_dict_str.items()])

        # Capture output to avoid cluttering console
        console.begin_capture()
        script_path, options_path = generate(
            experiment,
            machine=machine,
            num_processes=num_processes,
            max_time=max_time,
            output_dir=sweep_output_dir,
            link_name=link_name,
        )
        console.end_capture()
        shutil.copy(script_path, sweep_script_dir / script_path.name)
        script_path.unlink()
        script_path = sweep_script_dir / script_path.name

        # Link options file for ease of access
        (sweep_options_dir / options_path.name).symlink_to(options_path)

        scripts.append(script_path)
        print(f"[info]  Script written to: {script_path}[/]")
        print(f"[info]  Output directory:  {sweep_output_dir / link_name}[/]")

    print(f"[info]All scripts written to: {sweep_script_dir}[/]")
    print("")
    if dry_run:
        print(f"[info]Dry run mode: not submitting jobs.[/]")
        return
    print(f"[h2]Submitting jobs...[/]")
    for script_path in scripts:
        run(script_path)

    print(f"[success]All jobs submitted![/]")


def uq(
    experiment: ExperimentConfig,
    machine: Optional[Machine],
    num_processes: int,
    max_time: Optional[str] = None,
    parameters: dict = {},
    sweep_name: str = 'uq',
    yes: bool = False,
    dry_run: bool = False,
):
    """Generate flux scripts for a UQ study."""
    options = experiment.user_options.copy()
    num_runs = len(list(parameters.values())[0])
    scratch_dir = Path(config.get_fallback('SCRATCH_DIR'))
    output_dir = Path(config.get_fallback('OUTPUT_DIR', Path.cwd() / 'output'))
    if machine is None:
        machine = detect_machine()
        if machine is None:
            raise ValueError("Could not detect machine. Please specify a machine.")

    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    sweep_name = experiment.name + f'-{sweep_name}-' + '-'.join(parameters.keys())
    sweep_output_dir = (output_dir / sweep_name).resolve()
    sweep_script_dir = (sweep_output_dir / 'flux_scripts').resolve()
    sweep_options_dir = (sweep_output_dir / 'options').resolve()

    print(f'{experiment}')
    print("")
    print(f'[h2]Parameter Sweep[/]')
    print(f"  • Sweeping {len(parameters)} parameters:")
    for name, values in parameters.items():
        print(f"    • {name}: {', '.join(map(str, values))}")
    print(f"  • Total number of runs: {num_runs}")
    print(f"  • Number of processes per run: {num_processes}")
    print(f"  • Number of nodes per run: {int(ceil(num_processes / get_machine_config(machine).gpus_per_node))}")
    print(f"  • Output directory: {sweep_output_dir}")
    print(f"  • Script directory: {sweep_script_dir}")

    if sweep_output_dir.exists() and not yes:
        yn = console.input(f"[warning]Directory {sweep_output_dir} already exists. Remove?[/] (y/n) ")
        if yn.lower() != 'y':
            raise FileExistsError(f"Directory {sweep_output_dir} already exists.")
    if sweep_output_dir.exists():
        shutil.rmtree(sweep_output_dir)
    sweep_output_dir.mkdir(parents=True, exist_ok=True)
    sweep_script_dir.mkdir()
    sweep_options_dir.mkdir()

    scripts = []
    for params in zip(*parameters.values()):
        param_dict = dict(zip(parameters.keys(), params))
        param_dict_str = dict(zip(parameters.keys(), map(str, params)))
        new_options = options.copy()
        new_options.update(param_dict_str)
        experiment.user_options = new_options

        print(f"[info]Generating script for parameters:")
        for name, val in param_dict.items():
            print(f"[info]    • {name}: {val}")

        link_name = '---'.join([f"{key}-{value}" for key, value in param_dict_str.items()])

        # Capture output to avoid cluttering console
        console.begin_capture()
        script_path, options_path = generate(
            experiment,
            machine=machine,
            num_processes=num_processes,
            max_time=max_time,
            output_dir=sweep_output_dir,
            link_name=link_name,
        )
        console.end_capture()
        shutil.copy(script_path, sweep_script_dir / script_path.name)
        script_path.unlink()
        script_path = sweep_script_dir / script_path.name

        # Link options file for ease of access
        (sweep_options_dir / options_path.name).symlink_to(options_path)

        scripts.append(script_path)
        print(f"[info]  Script written to: {script_path}[/]")
        print(f"[info]  Output directory:  {sweep_output_dir / link_name}[/]")

    print(f"[info]All scripts written to: {sweep_script_dir}[/]")
    print("")
    if dry_run:
        print(f"[info]Dry run mode: not submitting jobs.[/]")
        return
    print(f"[h2]Submitting jobs...[/]")
    for script_path in scripts:
        run(script_path)

    print(f"[success]All jobs submitted![/]")



from enum import Enum
from dataclasses import dataclass, field
import platform
import getpass
import os
from pathlib import Path
from rich import print

from ..import config

detection_warning_printed = False


class GPUBackend(Enum):
    ROCM = 'rocm'
    CUDA = 'cuda'
    UNKNOWN = 'unknown'


@dataclass
class MachineConfig:
    gpus_per_node: int
    bank: str
    partition: str
    max_time: str
    ceed_backend: str
    parallel_filesystem: Path
    packages: list[str] = field(default_factory=list)
    defines: dict[str, str] = field(default_factory=dict)
    flux_args: list[str] = field(default_factory=list)
    petsc_args: list[str] = field(default_factory=list)


class Machine(Enum):
    """Enumeration of the machines that can be used for the experiments."""
    TUOLUMNE = 'tuolumne'
    TIOGA = 'tioga'
    LASSEN = 'lassen'
    DEFAULT = 'default'


def get_machine_config(machine: Machine) -> MachineConfig:
    """Get the configuration for the specified machine."""
    machine = Machine(machine)
    if machine == Machine.TUOLUMNE:
        tuo_packages = [
            'PrgEnv-amd',
            'rocmcc/6.4.3beta2-magic',
            'rocm/6.4.3beta2',
            'cray-mpich/9.0.1',
            'craype-accel-amd-gfx942',
            'cray-python',
            'cray-libsci_acc',
            'cray-hdf5-parallel/1.14.3.5',
            'flux_wrappers',
        ]
        tuo_defines = {
            'HSA_XNACK': '1',
            'MPICH_GPU_SUPPORT_ENABLED': '1',
            'MPICH_SMP_SINGLE_COPY_MODE': 'XPMEM',
        }
        gpu_mode = config.GPUMode(config.get_fallback('GPU_MODE', config.GPUMode.SPX))
        if gpu_mode == config.GPUMode.SPX:
            gpus_per_node = 4
        elif gpu_mode == config.GPUMode.CPX:
            gpus_per_node = 24
        else:  # gpu mode: TPX
            gpus_per_node = 12
        petsc_args = [
            '-vec_type hip',
            '-coarse_dm_mat_type aijhipsparse'
        ]
        return MachineConfig(
            gpus_per_node=gpus_per_node,
            bank='uco',
            partition='pbatch',
            max_time='12h',
            ceed_backend='/gpu/hip/gen',
            parallel_filesystem=Path('/p/lustre5'),
            packages=tuo_packages,
            defines=tuo_defines,
            flux_args=[f"--setattr=gpumode={gpu_mode}", "--conf=resource.rediscover=true"],
            petsc_args=petsc_args
        )
    elif machine == Machine.TIOGA:
        tioga_packages = [
            'rocmcc/6.4.0-cce-19.0.0d-magic',
            'rocm/6.4.0',
            'craype-accel-amd-gfx90a',
            'cray-python',
            'cray-libsci_acc',
            'cray-hdf5-parallel/1.14.3.5',
            'flux_wrappers',
            'cray-mpich/8.1.32',  # needed while 8.1.33 is in beta
        ]
        tioga_defines = {
            'HSA_XNACK': '1',
            'MPICH_GPU_SUPPORT_ENABLED': '1',
        }
        petsc_args = [
            '-vec_type hip',
            '-coarse_dm_mat_type aijhipsparse'
        ]
        return MachineConfig(
            gpus_per_node=8,
            bank='uco',
            partition='pdebug',
            max_time='12h',
            ceed_backend='/gpu/hip/gen',
            parallel_filesystem=Path('/p/lustre2'),
            packages=tioga_packages,
            defines=tioga_defines,
            petsc_args=petsc_args)
    elif machine == Machine.LASSEN:
        lassen_packages = [
            'clang/ibm-18.1.8-cuda-11.8.0-gcc-11.2.1',
            'cuda/11.8.0',
            'base-gcc/11.2.1',
            'essl',
            'lapack',
            'python/3.11.5',
        ]
        return MachineConfig(
            gpus_per_node=8,
            bank='uco',
            partition='pdebug',
            max_time='12h',
            ceed_backend='/gpu/cuda/gen',
            parallel_filesystem=Path('/p/gpfs1'),
            packages=lassen_packages)
    else:
        raise ValueError(f'Invalid machine: {machine}')


def detect_gpu_backend() -> GPUBackend:
    machine = detect_machine()
    if machine in [Machine.TUOLUMNE, Machine.TIOGA]:
        return GPUBackend.ROCM
    elif machine in [Machine.LASSEN]:
        return GPUBackend.CUDA

    if "CUDA_DIR" in os.environ.keys():
        return GPUBackend.CUDA
    if "ROCM_PATH" in os.environ.keys():
        return GPUBackend.ROCM
    return GPUBackend.UNKNOWN


def detect_machine() -> Machine | None:
    """Detect the machine that the script is running on."""
    global detection_warning_printed
    hostname = platform.node()
    if hostname.startswith('tuolumne'):
        return Machine.TUOLUMNE
    elif hostname.startswith('tioga'):
        return Machine.TIOGA
    elif hostname.startswith('lassen'):
        return Machine.LASSEN
    if not detection_warning_printed:
        print(f'[warning]Could not detect machine from hostname: {hostname}, are you connected to the right machine?')
        detection_warning_printed = True
    return None


def get_scratch(machine: Machine | None) -> Path | None:
    configured_dir = config.get('SCRATCH_DIR', machine=machine)
    if configured_dir or not machine or machine == Machine.DEFAULT:
        return Path(configured_dir) if configured_dir else None
    if machine and machine != Machine.DEFAULT:
        machine_config = get_machine_config(machine)
        default = machine_config.parallel_filesystem / getpass.getuser() / 'ratel-scratch'
        config.set('SCRATCH_DIR', f'{default}', machine=machine, quiet=False)
        return default



from abc import ABC, abstractmethod
from pathlib import Path
from enum import Enum
from typing import Optional


class LogViewType(Enum):
    FLAMEGRAPH = "flamegraph"
    XML = "xml"
    DETAIL = "detail"
    TEXT = "text"

    def to_petsc(self) -> str:
        match self:
            case self.FLAMEGRAPH:
                return ".txt:ascii_flamegraph"
            case self.XML:
                return ".xml:ascii_xml"
            case self.DETAIL:
                return ".py:ascii_info_detail"
            case self.TEXT:
                return ".txt"


class ExperimentConfig(ABC):
    _name: str
    _description: str
    _base_config: str
    _logview: Optional[LogViewType]

    def __init__(self, name: str, description: Optional[str], base_config: str, pretty_name: Optional[str] = None):
        self._name = name
        self._pretty_name = pretty_name if pretty_name is not None else name
        self._description = description or ''
        self._base_config = base_config
        self._logview = None
        self._user_options = dict()
        self.diagnostic_options = dict()

    @property
    def name(self) -> str:
        return self._name

    @property
    def pretty_name(self) -> str:
        return self._pretty_name

    @property
    def description(self) -> str:
        return self._description

    @property
    def base_config(self) -> str:
        return self._base_config

    @property
    @abstractmethod
    def mesh_options(self) -> str:
        raise NotImplementedError

    @property
    def logview(self) -> LogViewType | None:
        return self._logview

    @logview.setter
    def logview(self, value: Optional[LogViewType]):
        self._logview = LogViewType(value) if value is not None else None

    def parse_user_args(self, args) -> dict:
        options = {}
        i = 0
        while i < len(args):
            if args[i].startswith('-'):
                key = args[i].lstrip('-')
                if i + 1 < len(args) and not args[i + 1].startswith('-'):
                    value = args[i + 1]
                    i += 2
                else:
                    value = ''
                    i += 1
                options[key] = value
        return options

    @property
    def diagnostic_config(self):
        return '\n# Diagnostic output options\n' + '\n'.join([
            f"{key}: {value}"
            for key, value in self.diagnostic_options.items()
        ])

    @property
    def user_config(self) -> str:
        return '\n# User-provided options\n' + '\n'.join([
            f"{key}: {value}"
            for key, value in self._user_options.items()
        ])

    @property
    def user_options(self) -> dict:
        return self._user_options

    @user_options.setter
    def user_options(self, options: dict | list):
        if isinstance(options, list):
            self._user_options = self.parse_user_args(options)
        elif isinstance(options, dict):
            self._user_options = options
        else:
            raise TypeError("user_options must be a list or a dict")

    @property
    def config(self) -> str:
        config = self.base_config + self.mesh_options + self.diagnostic_config + self.user_config
        match self.logview:
            case LogViewType.FLAMEGRAPH | LogViewType.XML | LogViewType.DETAIL:
                config += '\n' + '\n'.join([
                    f'log_view: :log_view{self.logview.to_petsc()}',
                    'log_view_gpu_time:',
                ])
            case LogViewType.TEXT:
                config += '\n' + '\n'.join([
                    f'log_view: :log_view{self.logview.to_petsc()}',
                    'log_view_gpu_time:',
                    # 'log_view_memory:'
                ])
        return config

    def write_config(self, output_dir: Path) -> Path:
        """Write the configuration file for the experiment to output_dir."""
        config_path = output_dir / f'{self.name}.yaml'
        with config_path.open('w') as f:
            f.write(self.config)

        return config_path



from pathlib import Path
import rich
from rich.syntax import Syntax
from multiprocessing import cpu_count
from enum import Enum
import gmsh
import numpy as np
import typer
from typing import Annotated, Optional, ClassVar
from abc import ABC, abstractmethod
from types import SimpleNamespace
import pandas as pd

from ...helper.experiment import ExperimentConfig, LogViewType
from ...helper import config
from ...helper.flux import flux, machines
from ...helper.utilities import run_once, callback_is_set

from .. import local
from ..sweep import load_sweep_specification


console = rich.get_console()
print = console.print


class BoundaryType(Enum):
    SLIP = "slip"
    SLIP_CLAMPED_TOP = "slip_clamped_top"
    CONTACT = "contact"

    def snes_options(self) -> str:
        if self == BoundaryType.CONTACT:
            return '\n'.join([
                "# SNES options for contact boundary conditions",
                "snes:",
                "  monitor:",
                "  max_it: 15",
                "  rtol: 1e-6",
                "augmented_lagrangian_inner_snes:",
                "  linesearch:",
                "    type: bisection",
                "    monitor:",
                "  max_it: 20",
                "  monitor:",
                "  ksp:",
                "    ew:",
                "    ew_version: 3",
                "    ew_rtol0: 1e-4",
                "    ew_rtolmax: 1e-4",
                ""
            ])
        else:
            return '\n'.join([
                "# SNES options for slip/clamped boundary conditions",
                "snes:",
                "  linesearch:",
                "    type: bisection",
                "    monitor:",
                "  max_it: 20",
                "  monitor:",
                "  ksp:",
                "    ew:",
                "    ew_version: 3",
                "    ew_rtol0: 1e-4",
                "    ew_rtolmax: 1e-4",
                ""
            ])


class MaterialType(Enum):
    DAMAGE = "damage"
    NEO_HOOKEAN = "neo_hookean"
    MONOCLINIC = "monoclinic"
    TRICLINIC = "triclinic"


@run_once
def register_keys():
    keys = {
        'VOXEL_SIZE': config.ConfigKey('VOXEL_SIZE', 'Side length for each voxel', float),
        'LOAD_FRACTION': config.ConfigKey('LOAD_FRACTION', 'Desired Final height/Initial height ratio', float, config.CheckBounded(0, 1)),
        'VOXEL_DATA': config.ConfigKey('VOXEL_DATA', 'Path to voxel data file', Path),
        'CHARACTERISTIC_LENGTH': config.ConfigKey('CHARACTERISTIC_LENGTH', 'Desired characteristic length of background mesh', float),
        'GRAIN_IDS': config.ConfigKey('GRAIN_IDS', 'Range-based list of IDs corresponding to different grains', str),
    }
    for name, key in keys.items():
        config.add_key(name, key)


register_keys()

DIE_PIXEL_SIZE_IP01 = 8.434786e-3  # mm/pixel # 8.434786 um/pixel
# DIE_RADIUS = 2.550608716  # mm # 2550.608716 um, 5_000 / (8.434786 * 2) + 6 pixels
# DIE_HEIGHT = 8.51913386  # mm # 8519.13386 um, 1010 pixels
# DIE_CENTER = [2.65695759, 2.65695759, 0]  # [2656.95759, 2656.95759, 0] um, 315, 315, 0 pixels


def compute_die_stats(voxel_data, voxel_size: float, buf: int):
    with open(voxel_data) as f:
        line = f.readline()
    _, nx, ny, nz = map(int, line.split())
    radius = (5.000 / (voxel_size * 2) + buf) * voxel_size
    height = nz * voxel_size
    center = [nx / 2 * voxel_size, ny / 2 * voxel_size, 0]
    return radius, height, center


def set_diagnostic_options(experiment: ExperimentConfig, save_forces: int, save_strain_energy: int, save_swarm: int, save_solution: int,
                           save_diagnostics: int, save: bool) -> None:
    """
    Set diagnostic options for the experiment.

    :param experiment: The experiment configuration object.
    :param save_forces: Whether to save forces.
    :param save_swarm: Whether to save swarm data.
    :param save_solution: Whether to save solution data.
    :param save_diagnostics: Whether to save diagnostics.
    :param save_all: Whether to save all data.
    """
    if not save:
        return
    if save_forces > 0:
        experiment.diagnostic_options["ts_monitor_surface_force_per_face"] = "ascii:forces.csv"
        experiment.diagnostic_options["ts_monitor_surface_force_per_face_interval"] = f"{save_forces}"
    if save_strain_energy > 0:
        experiment.diagnostic_options["ts_monitor_strain_energy"] = "ascii:strain_energy.csv"
        experiment.diagnostic_options["ts_monitor_strain_energy_interval"] = f"{save_strain_energy}"
    if save_swarm > 0:
        experiment.diagnostic_options["ts_monitor_swarm_solution"] = "ascii:swarm.xmf"
        experiment.diagnostic_options["ts_monitor_swarm_fields"] = "J,volume,rho,material,model state,elastic parameters"
        experiment.diagnostic_options["ts_monitor_swarm_solution_interval"] = f"{save_swarm}"
    if save_solution > 0:
        experiment.diagnostic_options["ts_monitor_solution"] = r"cgns:solution_%06d.cgns"
        experiment.diagnostic_options["ts_monitor_solution_interval"] = f"{save_solution}"
    if save_diagnostics > 0:
        experiment.diagnostic_options["ts_monitor_output_fields"] = r"cgns:output_fields_%06d.cgns"
        experiment.diagnostic_options["ts_monitor_output_fields_interval"] = f"{save_diagnostics}"


def get_mesh(
    characteristic_length: float,
    voxel_data: Path,
    scratch_dir: Path,
    load_fraction: float,
    voxel_size: float,
    voxel_buf: int = 0,
    bc_type: BoundaryType = BoundaryType.SLIP_CLAMPED_TOP,
) -> str:
    """
    Get a mesh file for the given voxel data and characteristic length, generating if necessary.

    :param characteristic_length: The desired characteristic length for the mesh.
    :param voxel_data: Path to the voxel data file (e.g., CT scan).
    :param scratch_dir: Directory to store the generated mesh file.
    :param load_fraction: Fraction of the load to apply to the die (default is 0.4).
    :param voxel_size: Size of each voxel in the voxel data
    :param voxel_buf: Number of buffer voxels to add to the mesh
    :param bc_type: Type of boundary condition to apply.

    :return: A dictionary of mesh options for the experiment configuration.
    """
    radius, height, center = compute_die_stats(voxel_data, voxel_size, voxel_buf)
    mesh_file: Path = generate_mesh(characteristic_length, voxel_data, voxel_size, voxel_buf, scratch_dir)

    options: str = '\n'.join([
        "",
        "# Mesh options generated by press_common.generate_mesh",
        "mpm_voxel:",
        f"  filename: {voxel_data.resolve()}",
        f"  pixel_size: {voxel_size}",
        "",
        "dm_plex:",
        f"  filename: {mesh_file.resolve()}",
        "  cgns_parallel:",
        "  box_label:",
        "  dim: 3",
        "  simplex: 0",
        "",
        "remap:",
        f"  direction: z",
        f"  scale: {(1 - load_fraction)} # (1 - load_fraction) to match displacement",
        "",
    ])
    if bc_type == BoundaryType.CONTACT:
        options += '\n'.join([
            "bc:",
            "  allow_no_clamp:",
            "  contact: 1,2,3,4,5,6",
            "  # Bottom",
            "  contact_1:",
            "    shape: platen",
            "    normal: 0,0,1",
            f"    center: {center[0]},{center[1]},{center[2]}",
            "    penalty_min: 10",
            "    penalty_max: 5e4",
            "    penalty_scale: 4",
            "    type: augmented_lagrangian",
            "    friction:",
            "      type: coulomb",
            "      kinetic: 0.5",
            "      penalty_min: 10",
            "      penalty_max: 5e3",
            "      penalty_scale: 2",
            "  # Top, compressing 40%",
            "  contact_2:",
            "    shape: platen",
            "    penalty_min: 10",
            "    penalty_max: 5e4",
            "    penalty_scale: 4",
            "    normal: 0,0,-1",
            f"    center: {center[0]},{center[1]},{center[2]+height}",
            f"    distance: {load_fraction * height} # load_fraction * height",
            "    type: augmented_lagrangian",
            "    friction:",
            "      type: coulomb",
            "      kinetic: 0.5",
            "      penalty_min: 10",
            "      penalty_max: 5e3",
            "      penalty_scale: 2",
            "  # Outside",
            "",
        ])
        for i in range(3, 7):
            options += '\n'.join([
                f"  contact_{i}:",
                "    shape: cylinder",
                "    axis: 0,0,1",
                f"    radius: {radius}",
                f"    center: {center[0]},{center[1]},{center[2]}",
                "    inside:",
                "    penalty_min: 10",
                "    penalty_max: 5e4",
                "    penalty_scale: 4",
                "    type: augmented_lagrangian",
                "    friction:",
                "      type: coulomb",
                "      kinetic: 0.5",
                "      penalty_min: 10",
                "      penalty_max: 5e3",
                "      penalty_scale: 2",
                "",
            ])
    elif bc_type == BoundaryType.SLIP_CLAMPED_TOP:
        options += '\n'.join([
            "bc:",
            "  clamp: 1,2",
            "  # Clamped displacement for top and bottom",
            "  clamp_2:",
            f"    translate: 0,0,{-load_fraction * height} # -load_fraction * height",
            "  # Prevent x,y expansion beyond the die boundary",
            "  slip: 3,4,5,6",
            "  slip_3:",
            "    components: 0,1",
            "  slip_4:",
            "    components: 0,1",
            "  slip_5:",
            "    components: 0,1",
            "  slip_6:",
            "    components: 0,1",
            "",
        ])
    elif bc_type == BoundaryType.SLIP:
        options += '\n'.join([
            "bc:",
            "  slip: 1,2,3,4,5,6",
            "  # Allow x,y displacement for top and bottom, prescribe z displacement",
            "  slip_1:",
            "    components: 2",
            "  slip_2:",
            "    components: 2",
            f"    translate: {-load_fraction * height} # -load_fraction * height",
            "  slip_3:",
            "    components: 0,1",
            "  slip_4:",
            "    components: 0,1",
            "  slip_5:",
            "    components: 0,1",
            "  slip_6:",
            "    components: 0,1",
            "",
        ])

    print(f"[info]Generated mesh options:[/]")
    syntax = Syntax(options, "yaml")
    print(syntax)

    return options


def generate_mesh(characteristic_length: float, voxel_data: Path,
                  voxel_size: float, voxel_buf: int, scratch_dir: Path) -> Path:
    mesh_dir = scratch_dir / "meshes"
    if not mesh_dir.exists():
        mesh_dir.mkdir(parents=True, exist_ok=True)
    if not voxel_data.exists():
        raise FileNotFoundError(f"Voxel data {voxel_data} does not exist")
    mesh_file = mesh_dir / f"cylinder_{voxel_data.stem}_CL{characteristic_length}.cgns"
    if mesh_file.exists():
        console.print(f"[info]Using existing mesh [/]{mesh_file}")
        return mesh_file.resolve()

    element_order = 1
    radius, height, center = compute_die_stats(voxel_data, voxel_size, voxel_buf)
    square_radius = radius * np.sqrt(0.125)

    layers = int(np.ceil(height / characteristic_length) + 1)
    print(f"[info]Voxel Size: [/]{voxel_size}")
    print(f"[info]Center: [/]{center}")
    print(f"[info]Radius: [/]{radius}")
    print(f"[info]Height: [/]{height}")
    print(f"[info]Number of layers: [/]{layers}")
    print(f"[info]Element order: [/]{element_order}")
    print(f"[info]Characteristic length: [/]{characteristic_length}")

    EDGE = 1
    SURFACE = 2
    VOLUME = 3

    inner_rad = radius - 0.5 * square_radius * (1 + np.sqrt(2))
    outer_rad = radius - square_radius
    avg_rad = 0.5 * (inner_rad + outer_rad)
    NPTS_SQUARE = int(np.ceil(0.5 / characteristic_length * (square_radius + np.pi / 4 * radius))) + 1
    NPTS_RADIAL = int(np.ceil(avg_rad / characteristic_length)) + 1

    gmsh.initialize()

    # set mesh options
    gmsh.option.setNumber("General.NumThreads", cpu_count())
    gmsh.option.setNumber("Mesh.Algorithm3D", 10)
    gmsh.option.setNumber("Mesh.ElementOrder", element_order)
    gmsh.option.setNumber("Mesh.HighOrderOptimize", 1 if element_order > 1 else 0)
    gmsh.option.setNumber("Mesh.Binary", 1)

    p1 = gmsh.model.geo.addPoint(*center)
    p2 = gmsh.model.geo.addPoint(*(center + np.array([square_radius, 0, 0])))
    offset = (0.5 + 0.5 / np.sqrt(2))
    p3 = gmsh.model.geo.addPoint(*(center + np.array([square_radius * offset, square_radius * offset, 0])))
    p4 = gmsh.model.geo.addPoint(*(center + np.array([0, square_radius, 0])))

    p5 = gmsh.model.geo.addPoint(*(center + np.array([radius, 0, 0])))
    p6 = gmsh.model.geo.addPoint(*(center + np.array([radius / np.sqrt(2), radius / np.sqrt(2), 0])))
    p7 = gmsh.model.geo.addPoint(*(center + np.array([0, radius, 0])))

    l1 = gmsh.model.geo.addLine(p1, p2)
    l2 = gmsh.model.geo.addLine(p2, p3)
    l3 = gmsh.model.geo.addLine(p3, p4)
    l4 = gmsh.model.geo.addLine(p4, p1)

    l5 = gmsh.model.geo.addLine(p2, p5)
    l6 = gmsh.model.geo.addCircleArc(p5, p1, p6)
    l7 = gmsh.model.geo.addLine(p6, p3)
    l8 = gmsh.model.geo.addCircleArc(p6, p1, p7)
    l9 = gmsh.model.geo.addLine(p7, p4)

    cl1 = gmsh.model.geo.addCurveLoop((l1, l2, l3, l4))
    cl2 = gmsh.model.geo.addCurveLoop((l5, l6, l7, -l2))
    cl3 = gmsh.model.geo.addCurveLoop((l9, -l3, -l7, l8))

    s1 = gmsh.model.geo.addPlaneSurface([cl1])
    s2 = gmsh.model.geo.addPlaneSurface([cl2])
    s3 = gmsh.model.geo.addPlaneSurface([cl3])

    for c in [l5, l7, l9]:
        gmsh.model.geo.mesh.setTransfiniteCurve(c, NPTS_RADIAL)
    for c in [l1, l2, l3, l4, l6, l8]:
        gmsh.model.geo.mesh.setTransfiniteCurve(c, NPTS_SQUARE)
    for s in [s1, s2, s3]:
        gmsh.model.geo.mesh.setTransfiniteSurface(s)
        gmsh.model.geo.mesh.setRecombine(SURFACE, s)

    gmsh.model.geo.synchronize()

    def add_quadrant(orig, angle):
        quadrant = gmsh.model.geo.copy(orig)
        lines = [id for dim, id in quadrant if dim == EDGE]
        for l in lines:
            if l - lines[0] + 1 in [l5, l7, l9]:
                gmsh.model.geo.mesh.setTransfiniteCurve(l, NPTS_RADIAL)
            else:
                gmsh.model.geo.mesh.setTransfiniteCurve(l, NPTS_SQUARE)
        for s in filter(lambda x: x[0] == SURFACE, quadrant):
            gmsh.model.geo.mesh.setTransfiniteSurface(s[1])
            gmsh.model.geo.mesh.setRecombine(s[0], s[1])
        gmsh.model.geo.rotate(quadrant, center[0], center[1], center[2], 0, 0, 1, angle)
        return list(set(gmsh.model.getEntities()).intersection(quadrant))

    quadrant1 = gmsh.model.getEntities()
    add_quadrant(quadrant1, np.pi / 2)
    add_quadrant(quadrant1, np.pi)
    add_quadrant(quadrant1, 3 * np.pi / 2)

    gmsh.model.geo.removeAllDuplicates()
    gmsh.model.geo.synchronize()

    disk = gmsh.model.getEntities(SURFACE)

    # Extrude the disk along (0, 0, height) with the specified number of layers.
    # The 'recombine=True' option tells gmsh to recombine triangular faces into quads.
    gmsh.model.geo.extrude(disk, 0, 0, height, [layers], recombine=True)

    # Synchronize to update the model with the new entities.
    gmsh.model.geo.synchronize()

    print('About to generate mesh')

    # generate mesh
    gmsh.model.mesh.generate(3)

    node_tags, coords, parametric_coords = gmsh.model.mesh.getNodes()
    types, tags, nodes = gmsh.model.mesh.getElements(VOLUME)
    gmsh.finalize()

    gmsh.initialize()
    v1 = gmsh.model.addDiscreteEntity(VOLUME)
    gmsh.model.mesh.addNodes(VOLUME, v1, node_tags, coords, parametric_coords)
    gmsh.model.mesh.addElements(VOLUME, v1, types, tags, nodes)
    # save mesh
    gmsh.write(str(mesh_file))
    gmsh.finalize()

    if not mesh_file.exists():
        raise RuntimeError(f"Failed to generate mesh file: {mesh_file}")

    print(f"[info]Mesh saved to [/]{mesh_file}")
    return mesh_file


class PressExperiment(ExperimentConfig, ABC):
    """Die press experiment using voxelized CT data and a synthetic mesh, using sticky air for voids"""
    @property
    @abstractmethod
    def solver_config(self) -> str:
        pass

    @property
    @abstractmethod
    def material_config(self) -> str:
        pass

    base_name: ClassVar[str]

    def __init__(
            self,
            voxel_data: Path,
            characteristic_length: float,
            voxel_size: float,
            voxel_buf: int,
            load_fraction: float,
            bc_type: BoundaryType,
            material: MaterialType,
            seed: Optional[int],
            base_name: str | None = None,
            pretty_name: str | None = None,
            description: str | None = None,
    ):
        if not voxel_data.exists():
            raise FileNotFoundError(f"Voxel data {voxel_data} does not exist")
        if characteristic_length <= 0:
            raise ValueError(f"characteristic_length must be greater than 0, got {characteristic_length:f}")
        if load_fraction <= 0.0 or load_fraction > 1.0:
            raise ValueError(f"load_fraction must be in (0.0, 1.0], got {load_fraction}")
        self.voxel_data: Path = voxel_data
        self.characteristic_length: float = characteristic_length
        self.load_fraction: float = load_fraction
        self.bc_type = bc_type
        self.scratch_dir: Path = Path(config.get_fallback('SCRATCH_DIR')).resolve()
        self.voxel_size: float = voxel_size
        self.voxel_buf: int = voxel_buf
        self.material: MaterialType = material
        self.seed: int = seed or np.random.SeedSequence().generate_state(1)[0]
        base_config = self.solver_config + '\n' + self.bc_type.snes_options() + '\n' + self.material_config + '\n'
        base_config += f'mpm_grains_label_value: {config.get_fallback("GRAIN_IDS", "2")}\n'
        name = f"{base_name}-{voxel_data.stem}-{material.value}-CL{characteristic_length}-LF{load_fraction}-{bc_type.value}"
        pretty_name = pretty_name or "Ratel iMPM Press Experiment"
        description = description or self.__doc__
        super().__init__(name=name, pretty_name=pretty_name, description=description, base_config=base_config)

    @property
    def mesh_options(self) -> str:
        if hasattr(self, '_mesh_options'):
            return getattr(self, '_mesh_options')
        options = get_mesh(
            self.characteristic_length,
            self.voxel_data,
            self.scratch_dir,
            load_fraction=self.load_fraction,
            bc_type=self.bc_type,
            voxel_size=self.voxel_size,
            voxel_buf=self.voxel_buf
        )
        setattr(self, '_mesh_options', options)
        return options

    def __str__(self) -> str:
        output = '\n'.join([
            f'[h1]{self.pretty_name}[/]',
            f'{self.description}',
            f"\n[h2]Mesh Options[/]",
            f"  • Characteristic length: {self.characteristic_length}",
            f"  • Voxel data: {self.voxel_data}",
            f"  • Voxel size: {self.voxel_size}",
            f"  • Load fraction: {self.load_fraction}",
        ])
        if self.user_options:
            output += "\n[h2]User Options[/]\n"
            output += "\n".join([f"  • {key}: {value}" for key, value in self.user_options.items()])
        return output

    @classmethod
    def create_options_callback(cls):
        def options_callback(
            ctx: typer.Context,
            voxel_data: Annotated[Path, typer.Option('--voxel-data', '--data', default_factory=lambda: config.get("VOXEL_DATA"), callback=callback_is_set, help="Path to voxel dump file")],
            characteristic_length: Annotated[float, typer.Option(
                '--characteristic-length', '--cl', min=0, max=6, default_factory=lambda: config.get("CHARACTERISTIC_LENGTH"), callback=callback_is_set, help="Characteristic length of background mesh"
            )],
            voxel_size: Annotated[float, typer.Option('--voxel-size', '--size', default_factory=lambda: config.get("VOXEL_SIZE"), callback=callback_is_set, help="Voxel side length, should be constant for a given voxel dump file")],
            material: Annotated[MaterialType, typer.Option('--material', help="Material model to use")],
            load_fraction: Annotated[float, typer.Option(
                '--load-fraction', '--lf', min=0.0, max=1.0, help="Percent of total cylinder height to compress"
            )] = float(config.get_fallback("LOAD_FRACTION", 0.4)),
            voxel_buffer: Annotated[int, typer.Option(help="Number of buffer voxel widths to add to the mesh")] = 0,
            bc_type: Annotated[BoundaryType, typer.Option(
                help="Type of boundary condition to apply")] = BoundaryType.SLIP_CLAMPED_TOP,
            seed: Annotated[Optional[int], typer.Option(
                help="Random seed for any stochastic components, mainly for anisotropic materials")] = None,
            machine: Annotated[Optional[machines.Machine], typer.Option(
                help="HPC machine to generate flux scripts for")] = None,
            num_processes: Annotated[int, typer.Option("-n", min=1, help="Number of MPI processes")] = 1,
            log_view: Annotated[Optional[LogViewType], typer.Option(help="Type of log view profiling to use")] = None,
            save_forces: Annotated[int, typer.Option(
                min=0, help="Interval to save surface forces, or 0 to disable")] = 1,
            save_strain_energy: Annotated[int, typer.Option(
                min=0, help="Interval to save strain energy, or 0 to disable")] = 1,
            save_swarm: Annotated[int, typer.Option(min=0, help="Interval to save swarm data, or 0 to disable")] = 0,
            save_solution: Annotated[int, typer.Option(
                min=0, help="Interval to save mesh solution, or 0 to disable")] = 0,
            save_diagnostics: Annotated[int, typer.Option(
                min=0, help="Interval to save projected diagnostic quantities, or zero to disable")] = 0,
            save: Annotated[bool, typer.Option(
                help="Global flag to enable or disable writing diagnostics. If False, nothing will be written")] = True,
            checkpoint: Annotated[int, typer.Option(
                min=0, help="Interval to save checkpoint files for restarting runs, or zero to disable")] = 0,
            max_time: Annotated[Optional[str], typer.Option(
                "-t", "--max-time", help="Flux time specification for max job length.")] = None,
            max_restarts: Annotated[int, typer.Option(help="Number of restart jobs to enqueue", min=0)] = 0,
            dry_run: Annotated[bool, typer.Option(help="If true, only generate scripts and exit")] = False,
            yes: Annotated[bool, typer.Option('-y', help="Automatically accept any confirmation prompts")] = False,
        ):
            """Common options for press experiments

            Args:
                ctx (typer.Context): Typer application context
                voxel_data (Path, optional): Path to voxel dump file.
                characteristic_length (float, optional): Characteristic length of background mesh.
                voxel_size (float, optional): Voxel side length, should be constant for a given voxel dump file.
                load_fraction (float, optional): Percent of total cylinder height to compress.
                voxel_buffer (int, optional): Number of buffer voxel widths to add to the mesh.
                bc_type (BoundaryType, optional): Type of boundary condition to apply.
                machine (machines.Machine, optional): HPC machine to generate flux scripts for.
                num_processes (int, optional): Number of MPI processes.
                log_view (LogViewType, optional): Type of log view profiling to use.
                save_forces (int, optional): Interval to save surface forces, or 0 to disable.
                save_strain_energy (int, optional): Interval to save strain energy, or 0 to disable.
                save_swarm (int, optional): Interval to save swarm data, or 0 to disable.
                save_solution (int, optional): Interval to save mesh solution, or 0 to disable.
                save_diagnostics (int, optional): Interval to save projected diagnostic quantities, or zero to disable.
                save (bool, optional): Global flag to enable or disable writing diagnostics. If False, nothing will be written.
                checkpoint (int, optional): Interval to save checkpoint files for restarting runs, or zero to disable.
                max_time (str, optional): Flux time specification for max job length.
                max_restarts(int, optional): Number of restart jobs to enqueue.
                dry_run (bool, optional): If true, only generate scripts and exit.
                yes (bool, optional): Automatically accept any confirmation prompts.
            """
            if not ctx.obj:
                ctx.obj = SimpleNamespace()
            context = ctx.obj
            context.experiment = cls(
                voxel_data,
                characteristic_length,
                voxel_size,
                voxel_buf=voxel_buffer,
                load_fraction=load_fraction,
                bc_type=bc_type,
                material=material,
                seed=seed,
            )
            context.experiment.logview = log_view
            set_diagnostic_options(
                context.experiment,
                save_forces=save_forces,
                save_strain_energy=save_strain_energy,
                save_swarm=save_swarm,
                save_solution=save_solution,
                save_diagnostics=save_diagnostics,
                save=save,
            )
            context.machine = machine if machine is not None else machines.detect_machine()
            context.num_processes = num_processes
            context.checkpoint = checkpoint
            context.max_time = max_time
            context.max_restarts = max_restarts
            context.dry_run = dry_run
            context.yes = yes
        return options_callback

    @classmethod
    def create_app(cls):
        app = typer.Typer(name=cls.base_name, callback=cls.create_options_callback(), help=cls.__doc__)

        @app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
        def run(
            ctx: typer.Context,
            out: Optional[Path] = None,
        ):
            """Run the experiment in the current shell (blocking)

            Hint: See `ratel-runner mpm press [EXPERIMENT] --help` for all options
            """
            ctx.obj.experiment.user_options = ctx.args
            local.run(
                ctx.obj.experiment,
                num_processes=ctx.obj.num_processes,
                out=out,
                dry_run=ctx.obj.dry_run
            )

        @app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
        def flux_run(
            ctx: typer.Context,
        ):
            """Run the experiment using the Flux job scheduler

            Hint: See `ratel-runner mpm press [EXPERIMENT] --help` for all options
            """
            ctx.obj.experiment.user_options = ctx.args
            if ctx.obj.dry_run:
                script_file, _ = flux.generate(
                    ctx.obj.experiment,
                    machine=ctx.obj.machine,
                    num_processes=ctx.obj.num_processes,
                    max_time=ctx.obj.max_time,
                    checkpoint_interval=ctx.obj.checkpoint,
                )
                print(f"Generated script saved to", script_file)
                print("Dry run, exiting.")
            else:
                flux.submit_series(
                    ctx.obj.experiment,
                    machine=ctx.obj.machine,
                    num_processes=ctx.obj.num_processes,
                    max_time=ctx.obj.max_time,
                    checkpoint_interval=ctx.obj.checkpoint,
                    max_restarts=ctx.obj.max_restarts
                )

        @app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
        def flux_sweep(
            ctx: typer.Context,
            sweep_spec: Path,
        ):
            """Run a parameter sweep using the Flux job scheduler.

            Hint: See `ratel-runner mpm press [EXPERIMENT] --help` for all options
            """
            ctx.obj.experiment.user_options = ctx.args
            sweep_params = load_sweep_specification(ctx, sweep_spec, quiet=True)
            flux.sweep(
                ctx.obj.experiment,
                machine=ctx.obj.machine,
                num_processes=ctx.obj.num_processes,
                max_time=ctx.obj.max_time,
                parameters=sweep_params,
                sweep_name=sweep_spec.stem,
                yes=ctx.obj.yes,
                dry_run=ctx.obj.dry_run
            )

        @app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
        def flux_uq(
            ctx: typer.Context,
            uq_spec: Path,
        ):
            """Run a parameter sweep using the Flux job scheduler."""
            ctx.obj.experiment.user_options = ctx.args
            uq_params = pd.read_csv(uq_spec).to_dict(orient='list')
            flux.uq(
                ctx.obj.experiment,
                machine=ctx.obj.machine,
                num_processes=ctx.obj.num_processes,
                max_time=ctx.obj.max_time,
                parameters=uq_params,
                sweep_name=uq_spec.stem,
                yes=ctx.obj.yes,
                dry_run=ctx.obj.dry_run
            )

        @app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
        def flux_strong_scaling(
            ctx: typer.Context,
            num_processes: Annotated[list[int], typer.Argument(min=1)],
            num_steps: int = 5,
        ):
            """Run a parameter sweep using the Flux job scheduler.

            Hint: See `ratel-runner mpm press [EXPERIMENT] --help` for all options
            """
            ctx.obj.experiment._name = ctx.obj.experiment._name + "-scaling"
            ctx.obj.experiment.user_options = ctx.args + [
                "--preload",
                "--ts_max_steps", f"{num_steps}"
            ]
            for np in num_processes:
                if ctx.obj.dry_run:
                    script_file, _ = flux.generate(
                        ctx.obj.experiment,
                        machine=ctx.obj.machine,
                        num_processes=np,
                        max_time=ctx.obj.max_time,
                    )
                    print(f"Generated script saved to", script_file)
                    print("Dry run, exiting.")
                else:
                    flux.submit_series(
                        ctx.obj.experiment,
                        machine=ctx.obj.machine,
                        num_processes=np,
                        max_time=ctx.obj.max_time,
                    )
        return app



import subprocess
from rich import print
import typer
from typing import Annotated, Optional

from .git import Repository
from .build import app
from . import build_petsc, build_libceed
from .. import config

URI = "https://gitlab.com/micromorph/ratel.git"
CONFIG_DEFAULT = """# Ratel configuration file
OPT=-O3 -g -fno-signed-zeros -freciprocal-math -ffp-contract=fast -march=native -fPIC -Wno-pass-failed -fassociative-math -fno-math-errno
"""


def get_repository():
    """Get the repository."""
    repo = Repository(URI)
    if not repo.is_cloned():
        repo.clone()
    return repo


@app.command("ratel")
def build_ratel(
    branch: Optional[str] = None,
    force: Annotated[bool, typer.Option('-f', '--force')] = False,
    petsc_branch: Optional[str] = None,
    libceed_branch: Optional[str] = None
):
    """Build Ratel and its dependencies."""
    petsc_dir, petsc_arch = build_petsc.build_petsc(branch=petsc_branch, force=force)
    libceed_dir = build_libceed.build_libceed(branch=libceed_branch, force=force)

    print("[h1]Building Ratel[/h1]")

    repo = get_repository()

    if branch is None:
        branch = repo.branch
    else:
        repo.checkout(branch)

    if not repo.is_up_to_date():
        pull = force
        if not force:
            pull = typer.confirm(f"[info]Repository is not up to date. Pull latest changes from {branch}?")
        if pull:
            print("[info]Pulling latest changes...")
            repo.pull()
    else:
        print("[info]Repository is up to date.")

    # Copy the configuration file to the repository directory
    config_file = repo.dir / 'config.mk'
    if force or not config_file.exists():
        print("[info]Creating default configuration file.")
        config_str = CONFIG_DEFAULT + "\n".join([
            f"PETSC_DIR={petsc_dir}",
            f"PETSC_ARCH={petsc_arch}",
            f"CEED_DIR={libceed_dir}",
            ""
        ])
        config_file.write_text(config_str)

    # Run the make command
    if force:
        make_command = ["make", "-B", "-j"]
    else:
        make_command = ["make", "-j"]
    print("[info]Running make command:")
    print("  > ", " ".join(make_command))
    subprocess.run(make_command, cwd=repo.dir, check=True)

    config.set("RATEL_DIR", str(repo.dir))

    print("[success]Ratel build complete with ", f"RATEL_DIR={repo.dir}.")



import typing


def to_base(base: int, num: int) -> typing.List[int]:
    ...


def from_base(base: int, num: typing.Iterable[int]) -> int:
    ...


def chunk_sequence(data: bytes, size: int) -> typing.Iterator[bytes]:
    ...



from typing import Callable
import typer


def run_once(func: Callable) -> Callable:
    """Decorator to ensure a function is called exactly one time"""
    def wrapper(*args, **kwargs):
        if not hasattr(wrapper, '__called'):
            setattr(wrapper, '__result', func(*args, **kwargs))
            setattr(wrapper, '__called', True)
        return getattr(wrapper, '__result')
    return wrapper


def callback_is_set(value):
    if value is None:
        raise typer.BadParameter("Required CLI option")
    return value



import typer
import json
from pathlib import Path
from rich import print
import os
from typing import Annotated, Any, Callable
from dataclasses import dataclass
from difflib import get_close_matches
from rich.table import Table
from enum import Enum

from .flux import machines

__all__ = ['get_app_dir', 'get', 'set', 'unset', 'get_fallback', 'app']


_configuration = None


class GPUMode(Enum):
    SPX = 'SPX'
    CPX = 'CPX'
    TPX = 'TPX'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


@dataclass
class ConfigKey:
    name: str
    description: str
    type: type
    check: Callable[[Any], bool] = lambda _: True

    def __eq__(self, key):
        if isinstance(key, ConfigKey):
            return self.name == key.name
        return self.name == key

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class CheckBounded():
    def __init__(self, lower=None, upper=None, inclusive_lower=False, inclusive_upper=False):
        assert lower is not None or upper is not None, "One of upper or lower bounds must be set"
        self.lower = lower
        self.upper = upper
        self.inclusive_lower = inclusive_lower
        self.inclusive_upper = inclusive_upper

    def __str__(self) -> str:
        if self.lower is None:
            return f"must be <{('=' if self.inclusive_upper else '')} {self.upper}"
        if self.upper is None:
            return f"must be >{('=' if self.inclusive_lower else '')} {self.lower}"
        bracket_lower = "[" if self.inclusive_lower else "("
        bracket_upper = "]" if self.inclusive_upper else ")"
        return f"must be in the interval {bracket_lower}{self.lower}, {self.upper}{bracket_upper}"

    def __call__(self, value) -> bool:
        good = value is not None
        if good and self.lower is not None:
            good = good and (value > self.lower or (self.inclusive_lower and value == self.lower))
        if good and self.upper is not None:
            good = good and (value < self.upper or (self.inclusive_upper and value == self.upper))
        if not good:
            print(f'[error]value {value} {self}')
            return False
        return True


_KNOWN_KEYS = {
    "PETSC_DIR": ConfigKey("PETSC_DIR", "PETSc repository directory", Path),
    "PETSC_ARCH": ConfigKey("PETSC_ARCH", "PETSc arch name", Path),
    "PETSC_CONFIG": ConfigKey("PETSC_CONFIG", "User-specified configuration file for PETSc", Path),
    "LIBCEED_DIR": ConfigKey("LIBCEED_DIR", "libCEED repository directory", Path),
    "RATEL_DIR": ConfigKey("RATEL_DIR", "Ratel repository directory", Path),
    "SCRATCH_DIR": ConfigKey("SCRATCH_DIR", "Directory in which to write simulation outputs", Path),
    "OUTPUT_DIR": ConfigKey("OUTPUT_DIR", "Directory in which to place symlinks to output directories", Path),
    "GPU_MODE": ConfigKey("GPU_MODE", "GPU mode, only matters for Tuolumne", GPUMode),
}


def add_key(key_name: str, key: ConfigKey):
    if key_name in _KNOWN_KEYS.keys():
        print(f'[warn]{key_name} already registered, overwriting')
    _KNOWN_KEYS[key_name] = key


def _get_config(machine: machines.Machine | None = None):
    """
    Get the configuration file, creating it if it does not exist.
    """
    if machine is None:
        machine = machines.detect_machine()
    app_dir = get_app_dir()
    if machine is None or machine == machines.Machine.DEFAULT:
        config_file = app_dir / "config.json"
    else:
        config_file = app_dir / f'{machine.value.lower()}' / "config.json"
    if not config_file.exists():
        config_file.parent.mkdir(parents=True, exist_ok=True)
        config = {}
        config_file.write_text(json.dumps(config, indent=2))
    return config_file


def _get_runtime_config(machine: machines.Machine | None = None) -> dict:
    """
    Read/write values in the application configuration file.
    """
    if machine is None:
        machine = machines.detect_machine()
    global _configuration
    if _configuration is None:
        _configuration = dict()
    if _configuration.get(machine, None) is not None:
        return _configuration[machine]
    config_path = _get_config(machine)
    _configuration[machine] = json.loads(config_path.read_text())
    return _configuration[machine]


def _write_runtime_config(ctx: typer.Context, **kwargs):
    global _configuration
    if not _configuration:
        return
    for machine, config in _configuration.items():
        config_file = _get_config(machine)
        config_file.write_text(json.dumps(config, indent=2))


app = typer.Typer(callback=_get_runtime_config, result_callback=_write_runtime_config)


def get_app_dir():
    return Path(typer.get_app_dir("ratel-runner"))


def unset(key: str, machine: machines.Machine | None = None):
    """
    Remove a key from the runtime configuration.
    """
    config = _get_runtime_config(machine=machine)
    if key in config:
        del config[key]


@app.command('unset')
def unset_cmd(key: str, machine: machines.Machine | None = None):
    """
    Remove a key from the configuration file.
    """
    unset(key, machine=machine)


def set(key: str, value: str, machine: machines.Machine | None = None, quiet: bool = True):
    """
    Set a key-value pair in the runtime configuration.
    """
    if key not in _KNOWN_KEYS.keys():
        print(f"[warn]Unknown key:[/warn] {key}")
        similar = get_close_matches(key, _KNOWN_KEYS.keys())
        if len(similar):
            print(f"[warn]  Similar keys:[/warn] {' '.join(similar)}")
        raise typer.Exit(1)
    config_key = _KNOWN_KEYS[key]
    if not config_key.check(config_key.type(value)):
        raise typer.Abort(f'Error: Invalid value {value} for key {key}')
    config = _get_runtime_config(machine)
    if not quiet:
        if config_key.name in config:
            print(f"Key {key} already exists. Overwriting value (old: {config[key]})")
        print(f"Setting {key} to {value}")
    config[config_key.name] = str(config_key.type(value))


@app.command('set')
def set_cmd(key: str, value: str, machine: machines.Machine | None = None):
    """
    Set a key-value pair in the configuration file.
    """
    set(key, value, machine=machine, quiet=False)


def get(key: str, machine: machines.Machine | None = None):
    """
    Internal helper function to get the value of a key from the config file.
    """
    if key not in _KNOWN_KEYS.keys():
        print(f"[warn]Unknown key:[/warn] {key}")
        similar = get_close_matches(key, _KNOWN_KEYS.keys())
        if len(similar):
            print(f"[warn]  Similar keys:[/warn] {' '.join(similar)}")
        raise typer.Exit(1)
    config_key = _KNOWN_KEYS[key]
    config = _get_runtime_config(machine)
    value = config.get(key, None)
    return config_key.type(value) if value else None


@app.command('get')
def get_cmd(key: str, machine: machines.Machine | None = None):
    """
    Get the value of a key in the configuration file.
    """
    val = get(key, machine=machine)
    print(f"{key}: ", "Key not found." if val is None else val)


def list(machine: machines.Machine | None = None):
    """
    List all keys in the configuration file.
    """
    config = _get_runtime_config(machine)
    if len(config) == 0:
        print("[success]Configuration empty, use[/success] `config set` [success]to add configuration variables")
    else:
        table = Table("Name", "Description", "Options", "Value")
        for key, value in config.items():
            config_key = _KNOWN_KEYS.get(key, ConfigKey(f'[warn]{key}', "[warn]UNKNOWN KEY[/warn]", str))
            options = ", ".join(str(e) for e in config_key.type) if issubclass(config_key.type, Enum) else ""
            table.add_row(f'[bold underline]{config_key.name}', f"{config_key.description}", options, value)
        print(table)


@app.command('list')
def list_cmd(machine: machines.Machine | None = None):
    """
    List all keys and values in the configuration file.
    """
    list(machine=machine)


@app.command('copy')
def copy_cmd(src: machines.Machine, dst: Annotated[machines.Machine | None, typer.Argument()] = None):
    """
    Copy all configuration variables from one machine to another
    """
    config_from: dict = _get_runtime_config(src)
    config_to: dict = _get_runtime_config(dst)
    config_to.clear()
    for key, value in config_from.items():
        config_to[key] = value


def get_fallback(key: str, default=None, machine: machines.Machine | None = None):
    """
    Get the value of a key from the runtime config, environment variable, default provided, or raise an error.
    """
    value = get(key, machine=machine)
    if value is not None:
        return value
    value = os.environ.get(key, None)
    if value is not None:
        return value
    if default is not None:
        return default
    else:
        raise ValueError(
            f"{key} not set. Please set the {key} environment variable or use the config command to set it.")



from ._utils import chunk_sequence, from_base, to_base
from ._wordlist import index_to_word, word_to_index


def _block_to_indices(block):
    if len(block) > 4:
        raise ValueError("block too big")

    # `menmonicode` uses little-endian numbers.
    num = from_base(256, reversed(block))

    indices = list(reversed(to_base(1626, num)))

    # Pad the list of indices to the correct size.
    length = {
        1: 1,
        2: 2,
        3: 3,
        4: 3,
    }[len(block)]
    indices += [0] * (length - len(indices))

    # The third byte in a block slightly leaks into the third word.  A
    # different set of words is used for this case to distinguish it from the
    # four byte case.
    if len(block) == 3:
        indices[-1] += 1626

    return indices


def _block_to_words(block):
    for i in _block_to_indices(block):
        yield index_to_word(i)


def mnencode(data):
    """Encode a bytes object as an iterator of tuples of words.

    >>> list(mnencode(b"avocado"))
    [('bicycle', 'visible', 'robert'), ('cloud', 'unicorn', 'jet')]

    :param bytes data:
        The binary data to encode.
    :returns:
        A list of tuples of between one and three words from the wordlist.
    """
    if not isinstance(data, (bytes, bytearray)):
        raise TypeError((
            "expected bytes or bytearray, got {cls}"
        ).format(cls=type(data).__name__))

    for block in chunk_sequence(data, 4):
        yield tuple(_block_to_words(block))


def mnformat(data, word_separator="-", group_separator="--"):
    """Encode a byte array as a sequence of grouped words, formatted as a
    single string.

    >>> mnformat(b"cucumber")
    'paris-pearl-ultra--gentle-press-total'

    :param bytes data:
        The binary data to encode.
    :param str word_separator:
        String that should be used to separate words within a group.
    :param str word_separator:
        String that should be used to separate groups of words.
    :return str:
        The data as an sequence of grouped words.
    """
    return group_separator.join(
        word_separator.join(group) for group in mnencode(data)
    )


def _words_to_block(words):
    if not isinstance(words, tuple):
        raise TypeError("expected tuple of words")

    if len(words) == 0:
        raise ValueError("no words in block")

    if len(words) > 3:
        raise ValueError("too many words in block")

    try:
        indices = list(word_to_index(word) for word in words)
    except KeyError as e:
        raise ValueError("word not recognized") from e

    # Calculate length of block.
    # Both three byte and four byte blocks map to three words but can be
    # distinguished as a different word list is used to encode the last word
    # in the three byte case.
    length = {
        1: 1,
        2: 2,
        3: 3 if indices[-1] >= 1626 else 4,
    }[len(words)]

    if length == 3:
        indices[2] -= 1626

    # Check that words in the second word list don't appear anywhere else in
    # the block.
    for index in indices:
        if index > 1626:
            raise ValueError((
                "unexpected three byte word: {word!r}"
            ).format(word=index_to_word(index)))

    num = from_base(1626, reversed(indices))

    block = bytes(reversed(to_base(256, num)))

    # Pad to correct length.
    return block.ljust(length, b'\x00')


def mndecode(data):
    """Decode an iterator of tuples of words to get a byte array

    >>> mndecode([('turtle', 'special', 'recycle'), ('ferrari', 'album')])
    b'potato'

    :param data:
        An iterator of tuples of between one and three words from the wordlist
    :return bytes:
        A :class:`bytes` object containing the decoded data
    """
    return b''.join(_words_to_block(words) for words in data)


def mnparse(string, word_separator="-", group_separator="--"):
    """Decode a mnemonicode string into a byte array.

    >>> mnparse('scoop-limit-recycle--ferrari-album')
    b'tomato'

    :param str string:
        The string containing the mnemonicode encoded data.
    :param str word_separator:
        String used to separate individual words in a group.
    :param str group_separator:
        String used to separate groups of words representing four byte blocks.
    :return bytes:
        A :class:`bytes` object containing the decoded data
    """
    if not isinstance(string, str):
        raise TypeError((
            "expected string, got {cls}"
        ).format(cls=type(string).__name__))

    # Empty string is a valid input but ``"".split(...)`` does not return an
    # empty iterator so we need to special case it.
    if len(string) == 0:
        return b''

    return mndecode(
        tuple(group.split(word_separator))
        for group in string.split(group_separator)
    )


__all__ = ['mnencode', 'mnformat', 'mndecode', 'mnparse']



from pathlib import Path
import importlib.resources
from typing import ClassVar

from .press_common import PressExperiment, MaterialType


class PressNoAirExperiment(PressExperiment):
    """Die press experiment using voxelized CT data and a synthetic mesh"""
    @property
    def solver_config(self) -> str:
        if hasattr(self, '_solver_options'):
            return getattr(self, '_solver_options')
        options = (importlib.resources.files(__package__ or '') / 'yml' / 'common_solver.yml').read_text()
        setattr(self, '_solver_options', options)
        return getattr(self, '_solver_options')

    @property
    def material_config(self) -> str:
        if hasattr(self, '_material_options'):
            return getattr(self, '_material_options')
        options = (importlib.resources.files(__package__ or '') / 'yml' /
                   f'press_no_air_{self.material.value}.yml').read_text()
        if self.material == MaterialType.MONOCLINIC or self.material == MaterialType.TRICLINIC:
            options += f"\nmpm_grains_random_seed: {self.seed}\n"
        setattr(self, '_material_options', options)
        return getattr(self, '_material_options')

    base_name: ClassVar[str] = 'no-air'

    def __init__(self, *super_args, **super_kwargs):
        super().__init__(*super_args, **super_kwargs, base_name=Path(__file__).stem.replace('_', '-'),
                         pretty_name="Ratel iMPM Press Experiment, no air, background stabilization", description=self.__doc__)

    @property
    def mesh_options(self) -> str:
        if hasattr(self, '_mesh_options'):
            return getattr(self, '_mesh_options')
        options = super().mesh_options
        if self.material == MaterialType.DAMAGE:
            options += '\n' + '\n'.join([
                "# Specific options for no air die experiment",
                f"mpm_grains_characteristic_length: {self.characteristic_length * 4}",
                f"mpm_binder_characteristic_length: {self.characteristic_length * 4}",
                f"mpm_stabilization_background_stiffness_characteristic_length: {self.characteristic_length * 4}",
                "",
            ])
        setattr(self, '_mesh_options', options)
        return options


app = PressNoAirExperiment.create_app()



# This module is responsible for running experiments locally.
from pathlib import Path
from rich import print
import subprocess
import datetime
from typing import Optional

from ..helper import config
from ..helper.experiment import ExperimentConfig


def run(experiment: ExperimentConfig, num_processes: int = 1, ratel_dir: Optional[Path] = None,
        out: Optional[Path] = None, scratch_dir: Optional[Path] = None, dry_run: bool = False):
    """Run the experiment locally."""
    # Resolve paths
    if ratel_dir is None:
        ratel_dir = Path(config.get_fallback('RATEL_DIR')).resolve()
    if scratch_dir is None:
        scratch_dir = Path(config.get_fallback('SCRATCH_DIR')).resolve()
    output_dir = Path(config.get_fallback('OUTPUT_DIR', Path.cwd() / 'output')).resolve()
    output_dir.mkdir(parents=True, exist_ok=True)
    scratch_dir = Path(config.get_fallback('SCRATCH_DIR', scratch_dir)).resolve()
    if out is not None:
        run_dir = scratch_dir / 'output' / out
    else:
        run_dir = scratch_dir / 'output' / f"{experiment.name}-{datetime.datetime.now().strftime(r'%Y-%m-%d_%H-%M-%S')}"
    run_dir = run_dir.resolve()
    print(f'{experiment}')
    print("")
    print(f"[h2]Simulation Options[/]")
    print(f"  • Ratel path: {ratel_dir}")
    print(f"  • Output directory: {output_dir}")
    print(f"  • Run directory: {run_dir}")
    print(f"  • Number of processes: {num_processes}")
    print("")

    if run_dir.exists():
        for file in run_dir.glob("*"):
            file.unlink()
        run_dir.rmdir()
    run_dir.mkdir(parents=True, exist_ok=True)
    output_link = output_dir / run_dir.name
    if output_link.exists():
        output_link.unlink()
    output_link.symlink_to(run_dir, True)

    config_file = experiment.write_config(run_dir)
    out_file = run_dir / "stdout.txt"
    err_file = run_dir / "stderr.txt"
    ratel_exe = ratel_dir / 'bin' / 'ratel-quasistatic'

    options = [
        "-options_file", f"{config_file.resolve()}",
    ]

    if num_processes > 1:
        cmd_arr = ["mpirun", "-np", f"{num_processes}", f"{ratel_exe}", *options]
    else:
        cmd_arr = [f"{ratel_exe}", *options]
    print(f"\n[h1]Running experiment[/]\n")
    print(f"[info]Running:\n  > [/]{' '.join(cmd_arr)}")

    if dry_run:
        print("[success]Dry run, exiting[/]")
        return

    if not dry_run:
        with out_file.open('wb') as outfile, err_file.open('wb') as err:
            subprocess.run(cmd_arr, cwd=run_dir.resolve(), stdout=outfile, stderr=err)
    else:
        print(f"Command: {' '.join(cmd_arr)}")



import subprocess
from rich import print
from typing import Annotated, Optional
import typer

from .git import Repository
from .build import app
from .. import config
from ..flux.machines import detect_gpu_backend, GPUBackend

URI = "https://github.com/CEED/libCEED.git"
CONFIG_DEFAULT = """# libCEED configuration file
OPT=-O3 -g -fno-signed-zeros -freciprocal-math -ffp-contract=fast -march=native -fPIC -Wno-pass-failed -fassociative-math -fno-math-errno

CC=$(shell which mpicc)
CXX=$(shell which mpicxx)
FC=
"""
CONFIG_ROCM = """
ROCM_DIR=${ROCM_PATH}
"""
CONFIG_CUDA = """
CUDA_DIR=${ROCM_PATH}
"""


def get_repository():
    """Get the repository."""
    repo = Repository(URI)
    if not repo.is_cloned():
        repo.clone()
    return repo


@app.command("libceed")
def build_libceed(branch: Optional[str] = None, force: Annotated[bool, typer.Option('-f', '--force')] = False):
    """Build PETSc."""
    print("[h1]Building libCEED[/h1]")

    repo = get_repository()

    if branch is None:
        branch = repo.branch
    else:
        repo.checkout(branch)

    if not repo.is_up_to_date():
        pull = force
        if not force:
            pull = typer.confirm(f"[info]Repository is not up to date. Pull latest changes from {branch}?")
        if pull:
            print("[info]Pulling latest changes...")
            repo.pull()
    else:
        print("[info]Repository is up to date.")

    # Copy the configuration file to the repository directory
    config_file = repo.dir / 'config.mk'
    if force or not config_file.exists():
        print("[info]Creating default configuration file.")
        backend = detect_gpu_backend()
        if backend == GPUBackend.CUDA:
            config_file.write_text('\n'.join([CONFIG_DEFAULT, CONFIG_CUDA]))
        elif backend == GPUBackend.ROCM:
            config_file.write_text('\n'.join([CONFIG_DEFAULT, CONFIG_ROCM]))
        else:
            config_file.write_text(CONFIG_DEFAULT)

    # Run the make command
    if force:
        make_command = ["make", "-B", "-j", "lib"]
    else:
        make_command = ["make", "-j", "lib"]
    print("[info]Running make command:")
    print("  > ", " ".join(make_command))
    subprocess.run(make_command, cwd=repo.dir, check=True)

    # Save the directory
    config.set("LIBCEED_DIR", str(repo.dir))

    print("[success]libCEED build complete with", f"LIBCEED_DIR={repo.dir}.")
    return repo.dir



from dataclasses import dataclass
import yaml
import rich
import re
from pathlib import Path
from typing import Annotated
import typer

__doc__ = "Load and write sweep specifications for Ratel iMPM experiments"
__all__ = ['load_sweep_specification', 'write_sweep_specification']

console = rich.get_console()
print = console.print

app = typer.Typer(help=__doc__)

# https://stackoverflow.com/a/4703508
numeric_pattern = r'[-+]?(?:(?:\d*\.\d+)|(?:\d+\.?))(?:[Ee][+-]?\d+)?'
range_pattern = re.compile(rf'^({numeric_pattern}):({numeric_pattern}):(\d+)$')


@dataclass
class ParameterRange:
    """
    Class to hold a range of values for a parameter in the sweep specification.

    The range is defined by a start value, an end value, and a count of values.
    The step size is calculated as (end - start) / (count - 1).
    """
    start: float
    end: float
    count: int

    @property
    def values(self) -> list[float]:
        step = (self.end - self.start) / (self.count - 1)
        return [self.start + i * step for i in range(self.count)]


def range_representer(dumper: yaml.Dumper, data: ParameterRange) -> str:
    """
    String representation of a range in the format start:end:count.
    """
    return dumper.represent_scalar("!parameter_range", f"{data.start}:{data.end}:{data.count}")


def range_parser(loader: yaml.Loader, node: yaml.Node) -> ParameterRange:
    """
    Parse a string in the format start:end:count into a tuple of (start, end, count).
    """
    value = loader.construct_scalar(node)
    match = range_pattern.match(value)
    if not match:
        raise ValueError(f"Invalid range format: {value}")
    start, end, count = match.groups()
    return ParameterRange(float(start), float(end), int(count))


yaml.add_constructor('!parameter_range', range_parser)
yaml.add_implicit_resolver('!parameter_range', range_pattern)
yaml.add_representer(ParameterRange, range_representer)


@app.command('load')
def load_sweep_specification(ctx: typer.Context, path: Path, quiet: bool = False) -> dict[str, list]:
    """
    Load the sweep specification from a YAML file.

    The YAML file should contain a dictionary where keys are parameter names
    and values are either a list of values or a range in the format start:end:count.
    """
    with open(path, 'r') as file:
        data = yaml.full_load(file)
        if not isinstance(data, dict):
            raise ValueError(f"Sweep specification file {path} is not a valid YAML file")
    sweep_parameters = {}
    for parameter, values in data.items():
        if isinstance(values, list):
            sweep_parameters[parameter] = values
        elif isinstance(values, ParameterRange):
            sweep_parameters[parameter] = values.values
        elif isinstance(values, str):
            sweep_parameters[parameter] = [values]
        else:
            raise ValueError(f"Invalid value for parameter {parameter}: {values}")
    if not quiet:
        print(f"Loaded sweep specification from {path}:")
        for parameter, values in sweep_parameters.items():
            print(f"  {parameter}: {values}")
    return sweep_parameters


@app.command('write')
def write_sweep_specification(path: Path, parameters: list[str]):
    """
    Write the sweep specification to a YAML file.

    The parameters should be a dictionary where keys are parameter names
    and values are lists of values.
    """
    mapped_parameters = {}
    for parameter, values in zip(parameters[::2], parameters[1::2]):
        match = range_pattern.match(values)
        if match:
            start, end, count = match.groups()
            mapped_parameters[parameter] = ParameterRange(float(start), float(end), int(count))
        else:
            val_list = values.split(',')
            mapped_parameters[parameter] = val_list

    with open(path, 'w') as file:
        yaml.dump(mapped_parameters, file)



# -----BEGIN ORIGINAL LICENSE-----
# mn_wordlist.c
# Copyright (c) 2000  Oren Tirosh <oren@hishome.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----END ORIGINAL LICENSE-----

_WORDLIST = [
    "academy", "acrobat", "active", "actor", "adam", "admiral",
    "adrian", "africa", "agenda", "agent", "airline", "airport",
    "aladdin", "alarm", "alaska", "albert", "albino", "album",
    "alcohol", "alex", "algebra", "alibi", "alice", "alien",
    "alpha", "alpine", "amadeus", "amanda", "amazon", "amber",
    "america", "amigo", "analog", "anatomy", "angel", "animal",
    "antenna", "antonio", "apollo", "april", "archive", "arctic",
    "arizona", "arnold", "aroma", "arthur", "artist", "asia",
    "aspect", "aspirin", "athena", "athlete", "atlas", "audio",
    "august", "austria", "axiom", "aztec", "balance", "ballad",
    "banana", "bandit", "banjo", "barcode", "baron", "basic",
    "battery", "belgium", "berlin", "bermuda", "bernard", "bikini",
    "binary", "bingo", "biology", "block", "blonde", "bonus",
    "boris", "boston", "boxer", "brandy", "bravo", "brazil",
    "bronze", "brown", "bruce", "bruno", "burger", "burma",
    "cabinet", "cactus", "cafe", "cairo", "cake", "calypso",
    "camel", "camera", "campus", "canada", "canal", "cannon",
    "canoe", "cantina", "canvas", "canyon", "capital", "caramel",
    "caravan", "carbon", "cargo", "carlo", "carol", "carpet",
    "cartel", "casino", "castle", "castro", "catalog", "caviar",
    "cecilia", "cement", "center", "century", "ceramic", "chamber",
    "chance", "change", "chaos", "charlie", "charm", "charter",
    "chef", "chemist", "cherry", "chess", "chicago", "chicken",
    "chief", "china", "cigar", "cinema", "circus", "citizen",
    "city", "clara", "classic", "claudia", "clean", "client",
    "climax", "clinic", "clock", "club", "cobra", "coconut",
    "cola", "collect", "colombo", "colony", "color", "combat",
    "comedy", "comet", "command", "compact", "company", "complex",
    "concept", "concert", "connect", "consul", "contact", "context",
    "contour", "control", "convert", "copy", "corner", "corona",
    "correct", "cosmos", "couple", "courage", "cowboy", "craft",
    "crash", "credit", "cricket", "critic", "crown", "crystal",
    "cuba", "culture", "dallas", "dance", "daniel", "david",
    "decade", "decimal", "deliver", "delta", "deluxe", "demand",
    "demo", "denmark", "derby", "design", "detect", "develop",
    "diagram", "dialog", "diamond", "diana", "diego", "diesel",
    "diet", "digital", "dilemma", "diploma", "direct", "disco",
    "disney", "distant", "doctor", "dollar", "dominic", "domino",
    "donald", "dragon", "drama", "dublin", "duet", "dynamic",
    "east", "ecology", "economy", "edgar", "egypt", "elastic",
    "elegant", "element", "elite", "elvis", "email", "energy",
    "engine", "english", "episode", "equator", "escort", "ethnic",
    "europe", "everest", "evident", "exact", "example", "exit",
    "exotic", "export", "express", "extra", "fabric", "factor",
    "falcon", "family", "fantasy", "fashion", "fiber", "fiction",
    "fidel", "fiesta", "figure", "film", "filter", "final",
    "finance", "finish", "finland", "flash", "florida", "flower",
    "fluid", "flute", "focus", "ford", "forest", "formal",
    "format", "formula", "fortune", "forum", "fragile", "france",
    "frank", "friend", "frozen", "future", "gabriel", "galaxy",
    "gallery", "gamma", "garage", "garden", "garlic", "gemini",
    "general", "genetic", "genius", "germany", "global", "gloria",
    "golf", "gondola", "gong", "good", "gordon", "gorilla",
    "grand", "granite", "graph", "green", "group", "guide",
    "guitar", "guru", "hand", "happy", "harbor", "harmony",
    "harvard", "havana", "hawaii", "helena", "hello", "henry",
    "hilton", "history", "horizon", "hotel", "human", "humor",
    "icon", "idea", "igloo", "igor", "image", "impact",
    "import", "index", "india", "indigo", "input", "insect",
    "instant", "iris", "italian", "jacket", "jacob", "jaguar",
    "janet", "japan", "jargon", "jazz", "jeep", "john",
    "joker", "jordan", "jumbo", "june", "jungle", "junior",
    "jupiter", "karate", "karma", "kayak", "kermit", "kilo",
    "king", "koala", "korea", "labor", "lady", "lagoon",
    "laptop", "laser", "latin", "lava", "lecture", "left",
    "legal", "lemon", "level", "lexicon", "liberal", "libra",
    "limbo", "limit", "linda", "linear", "lion", "liquid",
    "liter", "little", "llama", "lobby", "lobster", "local",
    "logic", "logo", "lola", "london", "lotus", "lucas",
    "lunar", "machine", "macro", "madam", "madonna", "madrid",
    "maestro", "magic", "magnet", "magnum", "major", "mama",
    "mambo", "manager", "mango", "manila", "marco", "marina",
    "market", "mars", "martin", "marvin", "master", "matrix",
    "maximum", "media", "medical", "mega", "melody", "melon",
    "memo", "mental", "mentor", "menu", "mercury", "message",
    "metal", "meteor", "meter", "method", "metro", "mexico",
    "miami", "micro", "million", "mineral", "minimum", "minus",
    "minute", "miracle", "mirage", "miranda", "mister", "mixer",
    "mobile", "model", "modem", "modern", "modular", "moment",
    "monaco", "monica", "monitor", "mono", "monster", "montana",
    "morgan", "motel", "motif", "motor", "mozart", "multi",
    "museum", "music", "mustang", "natural", "neon", "nepal",
    "neptune", "nerve", "neutral", "nevada", "news", "ninja",
    "nirvana", "normal", "nova", "novel", "nuclear", "numeric",
    "nylon", "oasis", "object", "observe", "ocean", "octopus",
    "olivia", "olympic", "omega", "opera", "optic", "optimal",
    "orange", "orbit", "organic", "orient", "origin", "orlando",
    "oscar", "oxford", "oxygen", "ozone", "pablo", "pacific",
    "pagoda", "palace", "pamela", "panama", "panda", "panel",
    "panic", "paradox", "pardon", "paris", "parker", "parking",
    "parody", "partner", "passage", "passive", "pasta", "pastel",
    "patent", "patriot", "patrol", "patron", "pegasus", "pelican",
    "penguin", "pepper", "percent", "perfect", "perfume", "period",
    "permit", "person", "peru", "phone", "photo", "piano",
    "picasso", "picnic", "picture", "pigment", "pilgrim", "pilot",
    "pirate", "pixel", "pizza", "planet", "plasma", "plaster",
    "plastic", "plaza", "pocket", "poem", "poetic", "poker",
    "polaris", "police", "politic", "polo", "polygon", "pony",
    "popcorn", "popular", "postage", "postal", "precise", "prefix",
    "premium", "present", "price", "prince", "printer", "prism",
    "private", "product", "profile", "program", "project", "protect",
    "proton", "public", "pulse", "puma", "pyramid", "queen",
    "radar", "radio", "random", "rapid", "rebel", "record",
    "recycle", "reflex", "reform", "regard", "regular", "relax",
    "report", "reptile", "reverse", "ricardo", "ringo", "ritual",
    "robert", "robot", "rocket", "rodeo", "romeo", "royal",
    "russian", "safari", "salad", "salami", "salmon", "salon",
    "salute", "samba", "sandra", "santana", "sardine", "school",
    "screen", "script", "second", "secret", "section", "segment",
    "select", "seminar", "senator", "senior", "sensor", "serial",
    "service", "sheriff", "shock", "sierra", "signal", "silicon",
    "silver", "similar", "simon", "single", "siren", "slogan",
    "social", "soda", "solar", "solid", "solo", "sonic",
    "soviet", "special", "speed", "spiral", "spirit", "sport",
    "static", "station", "status", "stereo", "stone", "stop",
    "street", "strong", "student", "studio", "style", "subject",
    "sultan", "super", "susan", "sushi", "suzuki", "switch",
    "symbol", "system", "tactic", "tahiti", "talent", "tango",
    "tarzan", "taxi", "telex", "tempo", "tennis", "texas",
    "textile", "theory", "thermos", "tiger", "titanic", "tokyo",
    "tomato", "topic", "tornado", "toronto", "torpedo", "total",
    "totem", "tourist", "tractor", "traffic", "transit", "trapeze",
    "travel", "tribal", "trick", "trident", "trilogy", "tripod",
    "tropic", "trumpet", "tulip", "tuna", "turbo", "twist",
    "ultra", "uniform", "union", "uranium", "vacuum", "valid",
    "vampire", "vanilla", "vatican", "velvet", "ventura", "venus",
    "vertigo", "veteran", "victor", "video", "vienna", "viking",
    "village", "vincent", "violet", "violin", "virtual", "virus",
    "visa", "vision", "visitor", "visual", "vitamin", "viva",
    "vocal", "vodka", "volcano", "voltage", "volume", "voyage",
    "water", "weekend", "welcome", "western", "window", "winter",
    "wizard", "wolf", "world", "xray", "yankee", "yoga",
    "yogurt", "yoyo", "zebra", "zero", "zigzag", "zipper",
    "zodiac", "zoom", "abraham", "action", "address", "alabama",
    "alfred", "almond", "ammonia", "analyze", "annual", "answer",
    "apple", "arena", "armada", "arsenal", "atlanta", "atomic",
    "avenue", "average", "bagel", "baker", "ballet", "bambino",
    "bamboo", "barbara", "basket", "bazaar", "benefit", "bicycle",
    "bishop", "blitz", "bonjour", "bottle", "bridge", "british",
    "brother", "brush", "budget", "cabaret", "cadet", "candle",
    "capitan", "capsule", "career", "cartoon", "channel", "chapter",
    "cheese", "circle", "cobalt", "cockpit", "college", "compass",
    "comrade", "condor", "crimson", "cyclone", "darwin", "declare",
    "degree", "delete", "delphi", "denver", "desert", "divide",
    "dolby", "domain", "domingo", "double", "drink", "driver",
    "eagle", "earth", "echo", "eclipse", "editor", "educate",
    "edward", "effect", "electra", "emerald", "emotion", "empire",
    "empty", "escape", "eternal", "evening", "exhibit", "expand",
    "explore", "extreme", "ferrari", "first", "flag", "folio",
    "forget", "forward", "freedom", "fresh", "friday", "fuji",
    "galileo", "garcia", "genesis", "gold", "gravity", "habitat",
    "hamlet", "harlem", "helium", "holiday", "house", "hunter",
    "ibiza", "iceberg", "imagine", "infant", "isotope", "jackson",
    "jamaica", "jasmine", "java", "jessica", "judo", "kitchen",
    "lazarus", "letter", "license", "lithium", "loyal", "lucky",
    "magenta", "mailbox", "manual", "marble", "mary", "maxwell",
    "mayor", "milk", "monarch", "monday", "money", "morning",
    "mother", "mystery", "native", "nectar", "nelson", "network",
    "next", "nikita", "nobel", "nobody", "nominal", "norway",
    "nothing", "number", "october", "office", "oliver", "opinion",
    "option", "order", "outside", "package", "pancake", "pandora",
    "panther", "papa", "patient", "pattern", "pedro", "pencil",
    "people", "phantom", "philips", "pioneer", "pluto", "podium",
    "portal", "potato", "prize", "process", "protein", "proxy",
    "pump", "pupil", "python", "quality", "quarter", "quiet",
    "rabbit", "radical", "radius", "rainbow", "ralph", "ramirez",
    "ravioli", "raymond", "respect", "respond", "result", "resume",
    "retro", "richard", "right", "risk", "river", "roger",
    "roman", "rondo", "sabrina", "salary", "salsa", "sample",
    "samuel", "saturn", "savage", "scarlet", "scoop", "scorpio",
    "scratch", "scroll", "sector", "serpent", "shadow", "shampoo",
    "sharon", "sharp", "short", "shrink", "silence", "silk",
    "simple", "slang", "smart", "smoke", "snake", "society",
    "sonar", "sonata", "soprano", "source", "sparta", "sphere",
    "spider", "sponsor", "spring", "acid", "adios", "agatha",
    "alamo", "alert", "almanac", "aloha", "andrea", "anita",
    "arcade", "aurora", "avalon", "baby", "baggage", "balloon",
    "bank", "basil", "begin", "biscuit", "blue", "bombay",
    "brain", "brenda", "brigade", "cable", "carmen", "cello",
    "celtic", "chariot", "chrome", "citrus", "civil", "cloud",
    "common", "compare", "cool", "copper", "coral", "crater",
    "cubic", "cupid", "cycle", "depend", "door", "dream",
    "dynasty", "edison", "edition", "enigma", "equal", "eric",
    "event", "evita", "exodus", "extend", "famous", "farmer",
    "food", "fossil", "frog", "fruit", "geneva", "gentle",
    "george", "giant", "gilbert", "gossip", "gram", "greek",
    "grille", "hammer", "harvest", "hazard", "heaven", "herbert",
    "heroic", "hexagon", "husband", "immune", "inca", "inch",
    "initial", "isabel", "ivory", "jason", "jerome", "joel",
    "joshua", "journal", "judge", "juliet", "jump", "justice",
    "kimono", "kinetic", "leonid", "lima", "maze", "medusa",
    "member", "memphis", "michael", "miguel", "milan", "mile",
    "miller", "mimic", "mimosa", "mission", "monkey", "moral",
    "moses", "mouse", "nancy", "natasha", "nebula", "nickel",
    "nina", "noise", "orchid", "oregano", "origami", "orinoco",
    "orion", "othello", "paper", "paprika", "prelude", "prepare",
    "pretend", "profit", "promise", "provide", "puzzle", "remote",
    "repair", "reply", "rival", "riviera", "robin", "rose",
    "rover", "rudolf", "saga", "sahara", "scholar", "shelter",
    "ship", "shoe", "sigma", "sister", "sleep", "smile",
    "spain", "spark", "split", "spray", "square", "stadium",
    "star", "storm", "story", "strange", "stretch", "stuart",
    "subway", "sugar", "sulfur", "summer", "survive", "sweet",
    "swim", "table", "taboo", "target", "teacher", "telecom",
    "temple", "tibet", "ticket", "tina", "today", "toga",
    "tommy", "tower", "trivial", "tunnel", "turtle", "twin",
    "uncle", "unicorn", "unique", "update", "valery", "vega",
    "version", "voodoo", "warning", "william", "wonder", "year",
    "yellow", "young", "absent", "absorb", "accent", "alfonso",
    "alias", "ambient", "andy", "anvil", "appear", "apropos",
    "archer", "ariel", "armor", "arrow", "austin", "avatar",
    "axis", "baboon", "bahama", "bali", "balsa", "bazooka",
    "beach", "beast", "beatles", "beauty", "before", "benny",
    "betty", "between", "beyond", "billy", "bison", "blast",
    "bless", "bogart", "bonanza", "book", "border", "brave",
    "bread", "break", "broken", "bucket", "buenos", "buffalo",
    "bundle", "button", "buzzer", "byte", "caesar", "camilla",
    "canary", "candid", "carrot", "cave", "chant", "child",
    "choice", "chris", "cipher", "clarion", "clark", "clever",
    "cliff", "clone", "conan", "conduct", "congo", "content",
    "costume", "cotton", "cover", "crack", "current", "danube",
    "data", "decide", "desire", "detail", "dexter", "dinner",
    "dispute", "donor", "druid", "drum", "easy", "eddie",
    "enjoy", "enrico", "epoxy", "erosion", "except", "exile",
    "explain", "fame", "fast", "father", "felix", "field",
    "fiona", "fire", "fish", "flame", "flex", "flipper",
    "float", "flood", "floor", "forbid", "forever", "fractal",
    "frame", "freddie", "front", "fuel", "gallop", "game",
    "garbo", "gate", "gibson", "ginger", "giraffe", "gizmo",
    "glass", "goblin", "gopher", "grace", "gray", "gregory",
    "grid", "griffin", "ground", "guest", "gustav", "gyro",
    "hair", "halt", "harris", "heart", "heavy", "herman",
    "hippie", "hobby", "honey", "hope", "horse", "hostel",
    "hydro", "imitate", "info", "ingrid", "inside", "invent",
    "invest", "invite", "iron", "ivan", "james", "jester",
    "jimmy", "join", "joseph", "juice", "julius", "july",
    "justin", "kansas", "karl", "kevin", "kiwi", "ladder",
    "lake", "laura", "learn", "legacy", "legend", "lesson",
    "life", "light", "list", "locate", "lopez", "lorenzo",
    "love", "lunch", "malta", "mammal", "margo", "marion",
    "mask", "match", "mayday", "meaning", "mercy", "middle",
    "mike", "mirror", "modest", "morph", "morris", "nadia",
    "nato", "navy", "needle", "neuron", "never", "newton",
    "nice", "night", "nissan", "nitro", "nixon", "north",
    "oberon", "octavia", "ohio", "olga", "open", "opus",
    "orca", "oval", "owner", "page", "paint", "palma",
    "parade", "parent", "parole", "paul", "peace", "pearl",
    "perform", "phoenix", "phrase", "pierre", "pinball", "place",
    "plate", "plato", "plume", "pogo", "point", "polite",
    "polka", "poncho", "powder", "prague", "press", "presto",
    "pretty", "prime", "promo", "quasi", "quest", "quick",
    "quiz", "quota", "race", "rachel", "raja", "ranger",
    "region", "remark", "rent", "reward", "rhino", "ribbon",
    "rider", "road", "rodent", "round", "rubber", "ruby",
    "rufus", "sabine", "saddle", "sailor", "saint", "salt",
    "satire", "scale", "scuba", "season", "secure", "shake",
    "shallow", "shannon", "shave", "shelf", "sherman", "shine",
    "shirt", "side", "sinatra", "sincere", "size", "slalom",
    "slow", "small", "snow", "sofia", "song", "sound",
    "south", "speech", "spell", "spend", "spoon", "stage",
    "stamp", "stand", "state", "stella", "stick", "sting",
    "stock", "store", "sunday", "sunset", "support", "sweden",
    "swing", "tape", "think", "thomas", "tictac", "time",
    "toast", "tobacco", "tonight", "torch", "torso", "touch",
    "toyota", "trade", "tribune", "trinity", "triton", "truck",
    "trust", "type", "under", "unit", "urban", "urgent",
    "user", "value", "vendor", "venice", "verona", "vibrate",
    "virgo", "visible", "vista", "vital", "voice", "vortex",
    "waiter", "watch", "wave", "weather", "wedding", "wheel",
    "whiskey", "wisdom", "deal", "null", "nurse", "quebec",
    "reserve", "reunion", "roof", "singer", "verbal", "amen",
    "ego", "fax", "jet", "job", "rio", "ski",
    "yes"
]

_WORD_INDECES = {
    word: index for index, word in enumerate(_WORDLIST)
}


def index_to_word(index):
    return _WORDLIST[index]


def word_to_index(word):
    return _WORD_INDECES[word]



#!/usr/tce/bin/python3

if __name__ == '__main__':
    import sys
    import os
    import shutil
    from pathlib import Path
    sys.path.insert(0, os.path.abspath('config'))
    import configure

    mpi_dir = Path(shutil.which('mpicc'))
    mpi_dir = mpi_dir.resolve().parent.parent

    # I hate hardcoding this, make sure to update if packages change
    XLLIBDIR = '/usr/tce/packages/xl/xl-2023.06.28-cuda-11.8.0-gcc-11.2.1/alllibs'

    configure_options = [
        '--download-kokkos',
        '--download-kokkos-kernels',
        '--download-zlib',
        '--download-hypre',
        '--download-metis',
        '--download-parmetis',
        '--download-triangle',
        '--download-hdf5',
        '--download-cgns',
        f'--with-blaslapack-dir={os.environ["ESSLLIBDIR64"]}',
        '--download-cmake',
        '--with-cuda-arch=70',
        '--with-cuda=1',
        f'--with-cuda-dir={os.environ["CUDA_HOME"]}',
        '--with-debugging=0',
        '--with-fc=0',
        '--with-mpi=1',
        f'--with-mpi-dir={mpi_dir}',
        '--with-64-bit-indices',
        '--COPTFLAGS=-O3 -g -mcpu=native -Wno-pass-failed -fassociative-math -fno-math-errno -fno-omit-frame-pointer -ffp-contract=fast',
        '--CXXOPTFLAGS=-O3 -g -mcpu=native -Wno-pass-failed -fassociative-math -fno-math-errno -fno-omit-frame-pointer -ffp-contract=fast',
        '--FCOPTFLAGS=-O3 -g',
        '--CUDAOPTFLAGS=-O3 -g',
        '--CUDAPPFLAGS=-std=c++17',
        '--LIBS=-lesslsmp -llapackforessl -lxlfmath -lxlf90_r -lxlsmp -lm -lblas -llapack -lscalapack -lcblas -llapacke',
        f'--LDFLAGS=-L{XLLIBDIR} -Wl,-rpath,{XLLIBDIR} -L{os.environ["ESSLLIBDIR64"]} ' +
        f'-Wl,-rpath,{os.environ["ESSLLIBDIR64"]} -L{os.environ["LAPACK_DIR"]}  -Wl,-rpath,{os.environ["LAPACK_DIR"]}',
        # f'--CFLAGS+=-L{XLLIBDIR} -Wl,-rpath,{XLLIBDIR} -L{os.environ["ESSLLIBDIR64"]} -Wl,-rpath,{os.environ["ESSLLIBDIR64"]} -L{os.environ["LAPACK_DIR"]}  -Wl,-rpath,{os.environ["LAPACK_DIR"]}',
        # f'--CXXFLAGS+=-L{XLLIBDIR} -Wl,-rpath,{XLLIBDIR} -L{os.environ["ESSLLIBDIR64"]} -Wl,-rpath,{os.environ["ESSLLIBDIR64"]} -L{os.environ["LAPACK_DIR"]}  -Wl,-rpath,{os.environ["LAPACK_DIR"]}',
        '--PETSC_ARCH=arch-lassen-kokkos-opt',
    ]
    configure.petsc_configure(configure_options)



from pathlib import Path
import typer
from typing import Annotated, Optional

from ...helper.flux import machines, flux
from ...helper.experiment import ExperimentConfig, LogViewType


_base_side_length = 0.1
_efficiency_base_config = f"""# Single material, simple test to determine optimal point/GPU
method: mpm
mpm:
  point_location_type: uniform

material: cube
cube:
  model: elasticity-mpm-neo-hookean-damage-current
  E: 50e2
  nu: 0.4
  rho: 15
  fracture_toughness: 1e3 # increase to reduce damage accumulation
  characteristic_length: 1 # 4*mesh_cl
  residual_stiffness: 1e-2
  damage_scaling: 1
  use_AT1: false


dm_plex:
  # Mesh specification determined by script
  dim: 3
  simplex: 0
  # Use spatial hashing for point location
  hash_location:

bc:
  # Clamp all sides
  clamp: 1,2,3,4,5,6

forcing: body
forcing_body_vector: -9.81,0,0

# TS
ts:
  dt: 0.001
  max_time: 0.01
  monitor:

# SNES
snes:
  type: newtonls
  linesearch:
    type: cp
    max_it: 10

# KSP
ksp:
  type: bcgsl
  norm_type: preconditioned

# Preconditioner
pc:
  type: jacobi
"""


class EfficiencyForcingExperiment(ExperimentConfig):
    """Efficiency experiment configurations for optimal point/GPU determination, no mesh remapping."""
    num_cells_1d: int

    def __init__(self, num_cells_1d: int, order: int = 1, points_per_cell: int = 27):
        self.num_cells_1d = num_cells_1d
        self.num_points_per_cell = points_per_cell
        self.order = order
        super().__init__(Path(__file__).stem.replace('_', '-'), self.__doc__, _efficiency_base_config)

    @property
    def mesh_options(self) -> str:
        # Scale the mesh size based on the number of faces
        side_length = _base_side_length * self.num_cells_1d
        mesh_options = '\n'.join([
            "",
            "# Experiment options for efficiency experiment",
            f"mpm_num_points_per_cell: {self.num_points_per_cell}",
            f"order: {self.order}",
            "",
            "# Mesh options generated by script:",
            "dm_plex:",
            f"  box_faces: {self.num_cells_1d},{self.num_cells_1d},{self.num_cells_1d}",
            f"  box_upper: {side_length},{side_length},{side_length}",
            "",
        ])
        return mesh_options

    @property
    def name(self) -> str:
        return f"{self._name}_{self.num_cells_1d}"

    def __str__(self) -> str:
        output = '\n'.join([
            f'[h1]Ratel iMPM Efficiency Experiment[/]',
            f'{self.description}',
            f"\n[h2]Mesh Options[/]",
            f"  • Solution basis order: {self.order}",
            f"  • Number of cells: {self.num_cells_1d}^3={self.num_cells_1d**3}",
            f"  • Points per cell: {self.num_points_per_cell}",
            f"  • Points per GPU: {self.num_points_per_cell * self.num_cells_1d**3}",
        ])
        if self.user_options:
            output += "\n[h2]User Options[/]\n"
            output += "\n".join([f"  • {key}: {value}" for key, value in self.user_options.items()])
        return output


__doc__ = EfficiencyForcingExperiment.__doc__

app = typer.Typer(name="efficiency", help=__doc__)


@app.command()
def write_config(num_cells_1d: Annotated[int, typer.Argument(min=3)],
                 output_dir: Path, log_view: Optional[LogViewType] = None):
    """Generate the efficiency experiment configuration."""
    experiment = EfficiencyForcingExperiment(num_cells_1d)
    experiment.logview = log_view
    experiment.write_config(output_dir)


@app.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True}
)
def flux_run(
    ctx: typer.Context,
    num_cells_1d: Annotated[int, typer.Argument(min=3)],
    order: Annotated[int, typer.Argument(min=1)] = 1,
    points_per_cell: Annotated[int, typer.Argument(min=8)] = 27,
    num_processes: Annotated[int, typer.Option("-n", min=1)] = 1,
    max_time: Annotated[str, typer.Option('-t')] = '1h',
    log_view: Optional[LogViewType] = None,
    machine: Optional[machines.Machine] = None,
    dry_run: bool = False
):
    """Run the efficiency experiment using flux."""
    experiment = EfficiencyForcingExperiment(num_cells_1d, order, points_per_cell)
    experiment.user_options = ctx.args
    experiment.logview = log_view
    script_file, _ = flux.generate(
        experiment,
        machine=machine,
        num_processes=num_processes,
        max_time=max_time,
    )
    if not dry_run:
        flux.run(script_file)



from pathlib import Path
import importlib.resources
from typing import ClassVar

from .press_common import PressExperiment


class PressStickyAirExperiment(PressExperiment):
    """Die press experiment using voxelized CT data and a synthetic mesh, using sticky air for voids"""
    @property
    def solver_config(self) -> str:
        if hasattr(self, '_solver_options'):
            return getattr(self, '_solver_options')
        options = (importlib.resources.files(__package__ or '') / 'yml' / 'common_solver.yml').read_text()
        setattr(self, '_solver_options', options)
        return getattr(self, '_solver_options')

    @property
    def material_config(self) -> str:
        if hasattr(self, '_material_options'):
            return getattr(self, '_material_options')
        options = (importlib.resources.files(__package__ or '') / 'yml' / 'press_sticky_air.yml').read_text()
        setattr(self, '_material_options', options)
        return getattr(self, '_material_options')

    base_name: ClassVar[str] = 'sticky-air'

    def __init__(self, *super_args, **super_kwargs,):
        super().__init__(*super_args, **super_kwargs, base_name=Path(__file__).stem.replace('_', '-'),
                         pretty_name="Ratel iMPM Press Experiment, sticky air", description=self.__doc__)

    @property
    def mesh_options(self) -> str:
        if hasattr(self, '_mesh_options'):
            return getattr(self, '_mesh_options')
        options = super().mesh_options
        options += '\n' + '\n'.join([
            "# Specific options for sticky air die experiment",
            f"mpm_void_characteristic_length: {self.characteristic_length * 4}",
            f"mpm_grains_characteristic_length: {self.characteristic_length * 4}",
            f"mpm_binder_characteristic_length: {self.characteristic_length * 4}",
            "",
        ])
        setattr(self, '_mesh_options', options)
        return options


app = PressStickyAirExperiment.create_app()



from pathlib import Path
import subprocess
import urllib
import urllib.parse
from rich import print
from ..flux.machines import detect_machine, get_scratch
import typer

from .. import config


class Repository:
    """Class representing a repository."""

    def __init__(self, uri: str):
        self.name: str = uri.split('/')[-1].split('.')[0]
        self.uri: str = uri
        if self.uri.startswith('git@'):
            self.uri_end: str = self.uri.split(':')[-1]
        else:
            self.uri_end: str = urllib.parse.urlsplit(self.uri).path[1:].removesuffix('.git')
        self.config_key: str = f'{self.name.upper()}_DIR'
        self._branch = None
        dir = config.get_fallback(self.config_key, "")
        if dir == "":
            machine = detect_machine()
            scratch = get_scratch(machine)
            base_path = Path(config.get_fallback('SCRATCH_DIR', f'{scratch}')).resolve() / 'build'
            if not base_path.exists():
                base_path.mkdir(parents=True)
            self.dir = base_path / self.name
        else:
            self.dir = Path(dir).resolve()

    def __str__(self):
        return f"Repository(name={self.name}, uri={self.uri}, uri_end={self.uri_end}, config_key={self.config_key}, dir={self.dir})"

    @property
    def branch(self):
        if self._branch:
            return self._branch
        result = subprocess.run(['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                                cwd=self.dir, check=True, capture_output=True)
        self._branch = result.stdout.decode().strip()
        return self._branch

    def checkout(self, branch):
        if self.branch == branch:
            return
        self.pull(branch=branch)

    def is_up_to_date(self):
        """Check if the repository is up to date."""
        try:
            subprocess.run(['git', 'fetch'], cwd=self.dir, check=True)
            result = subprocess.run(['git', 'status'], cwd=self.dir, check=True, capture_output=True)
            return "Your branch is up to date" in result.stdout.decode()
        except subprocess.CalledProcessError:
            return False

    def is_cloned(self):
        """Check if the repository is cloned."""
        if not self.dir.exists() or not (self.dir / '.git').exists():
            return False
        result = subprocess.run(['git', 'remote', 'show', 'origin'], cwd=self.dir, check=True, capture_output=True)
        if result.returncode != 0:
            return False
        result = result.stdout.decode()
        return self.uri_end in result

    def clone(self):
        """Clone the repository."""
        subprocess.run(['git', 'clone', self.uri, f"{self.dir}"], check=True)

    def pull(self, branch='main'):
        """Pull the latest changes from the repository."""
        subprocess.run(['git', 'fetch'], cwd=self.dir, check=True)
        res = subprocess.run(['git', 'checkout', branch], cwd=self.dir, check=False, capture_output=True)
        if res.returncode != 0:
            print(res.stderr.decode())
            if 'Your local changes to the following files would be overwritten by checkout' in res.stderr.decode():
                if typer.confirm('You have uncommitted changes. Stash and checkout?'):
                    subprocess.run(['git', 'stash'], cwd=self.dir, check=True)
                elif typer.confirm('Okay, would you rather reset changes? (THIS CANNOT BE UNDONE)'):
                    subprocess.run(['git', 'reset', '--hard', f'origin/{self.branch}'], cwd=self.dir, check=True)
                else:
                    print(
                        f'[error]Failed to checkout {branch} due to dirty git working tree. Commit, stash, or remove your changes to continue.')
                    raise typer.Exit(1)
                subprocess.run(['git', 'checkout', branch], cwd=self.dir, check=True)

            else:
                typer.Exit(1)
        self._branch = branch
        subprocess.run(['git', 'pull', '--ff'], cwd=self.dir, check=True)
